"""
this module provides the heatmap ploting engine, it supports both virtual time (fixed number of trace requests) and
real time, under both modes, it support using multiprocessing to do the plotting
"""

import pickle
import sys
from multiprocessing import Process, Queue, Pool

import numpy as np
from matplotlib import pyplot as plt

from mimircache.cacheReader.csvReader import csvCacheReader
from mimircache.cacheReader.plainReader import plainCacheReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader
from mimircache.oldModule.pardaProfiler import pardaProfiler


def prepare_heatmap_dat(bin_size=1000):
    reader = csvCacheReader("../data/trace_CloudPhysics_txt", 4, delimiter=' ')
    p = pardaProfiler(30000, reader)
    c_reuse_dist_long_array = p.get_reuse_distance()

    array_len = len(c_reuse_dist_long_array) // bin_size
    result = np.empty((array_len - 1, array_len), dtype=np.float32)
    result[:] = np.nan
    # (x, y) -> heat, x, y is the left, lower point of heat square
    # (x,y) means from time x to time y

    print(len(c_reuse_dist_long_array))
    # print((c_reuse_dist_long_array))


    with open('reuse.dat', 'w') as ofile:
        for i in c_reuse_dist_long_array:
            print(i)
            ofile.write(str(i) + '\n')
    sys.exit(1)

    for (x, y) in _get_xy_distribution(bin_size, len(c_reuse_dist_long_array)):
        print('({},{})'.format(x, y))
        hr = _calc_hit_rate(c_reuse_dist_long_array, cache_size, x, y)
        result[x // bin_size][y // bin_size] = hr

    print(result)
    with open('temp2', 'wb') as ofile:
        pickle.dump(result, ofile)


def prepare_heatmap_dat_multiprocess(bin_size=1000, cache_size=2000, num_of_process=8):
    reader = plainCacheReader("../data/parda.trace")
    total_line = reader.get_num_of_total_requests()
    p = pardaProfiler(30000, reader)
    c_reuse_dist_long_array = p.get_reuse_distance()

    array_len = len(c_reuse_dist_long_array) // bin_size
    result = np.empty((array_len - 1, array_len), dtype=np.float32)
    result[:] = np.nan
    # (x, y) -> heat, x, y is the left, lower point of heat square
    # (x,y) means from time x to time y

    print(len(c_reuse_dist_long_array))
    # print((c_reuse_dist_long_array[-2]))



    # for i in c_reuse_dist_long_array:
    #     print(i)

    l = _get_xy_distribution_list(bin_size, len(c_reuse_dist_long_array))

    divided_len = len(l) // num_of_process
    q = Queue()
    process_list = []
    for i in range(num_of_process):
        if i != num_of_process - 1:
            p = Process(target=_calc_hit_rate_multiprocess,
                        args=(c_reuse_dist_long_array, cache_size, l[divided_len * i:divided_len * (i + 1)], q))
            process_list.append(p)
            p.start()
        else:
            p = Process(target=_calc_hit_rate_multiprocess,
                        args=(c_reuse_dist_long_array, cache_size, l[divided_len * i:len(l)], q))
            process_list.append(p)
            p.start()

    num_left = len(l)
    while num_left:
        t = q.get()
        result[t[0] // bin_size][t[1] // bin_size] = t[2]
        num_left -= 1
        print(num_left)

    print(result)
    with open('temp', 'wb') as ofile:
        pickle.dump(result, ofile)


def prepare_heatmap_dat_multiprocess_ts(datapath="../data/trace_CloudPhysics_bin",
                                        time_interval=10000000, cache_size=2000, num_of_process=8, calculate=True):
    reader = vscsiCacheReader(datapath)
    reuse_dist_python_list = []
    break_points = None

    if calculate:
        p = pardaProfiler(30000, reader)
        c_reuse_dist_long_array = p.get_reuse_distance()

        # print(len(c_reuse_dist_long_array))
        for i in c_reuse_dist_long_array:
            reuse_dist_python_list.append(i)

        with open('reuse.dat', 'wb') as ofile:
            pickle.dump(reuse_dist_python_list, ofile)
        print(len(reuse_dist_python_list))

    else:
        with open('reuse.dat', 'rb') as ifile:
            reuse_dist_python_list = pickle.load(ifile)

        if os.path.exists('break_points_' + str(time_interval) + '.dat'):
            with open('break_points_' + str(time_interval) + '.dat', 'rb') as ifile:
                break_points = pickle.load(ifile)

    # xy_list is a two dimensional square matrix, the returned size is the size of one dimension

    # old
    # xy_list, xy_dict, size = _get_xy_distribution_list_timestamp(reader, time_interval)

    # new
    if not break_points:
        break_points = _get_xy_distribution_list_timestamp(reader, time_interval)
        with open('break_points_' + str(time_interval) + '.dat', 'wb') as ifile:
            pickle.dump(break_points, ifile)

    # old
    # array_len = size

    # new
    array_len = len(break_points)
    result = np.empty((array_len, array_len), dtype=np.float32)
    result[:] = np.nan
    # (x, y) -> heat, x, y is the left, lower point of heat square
    # (x,y) means from time x to time y


    # divided_len = len(xy_list)//num_of_process      # old
    q = Queue()
    process_list = []

    # efficiency can be further improved by porting into Cython and improve parallel logic
    map_list = [i for i in range(len(break_points) - 1)]
    print(len(map_list))

    count = 0
    async_result_list = []
    with Pool(processes=num_of_process, initializer=_cal_hit_rate_multiprocess_ts2_init,
              initargs=[reuse_dist_python_list, cache_size, break_points, q]) as p:
        for l in p.imap_unordered(_calc_hit_rate_multiprocess_ts2, map_list):
            count += 1
            print("%2.2f%%" % (count / len(map_list) * 100))
            for t in l:  # l is a list of (x, y, hr)
                result[t[0]][t[1]] = t[2]

    # old version not efficient
    # for i in range(num_of_process):
    #     if i!=num_of_process-1:
    #         p = Process(target=_calc_hit_rate_multiprocess_ts,
    #                     args=(reuse_dist_python_list, cache_size, xy_list[divided_len*i:divided_len*(i+1)], q))
    #         process_list.append(p)
    #         p.start()
    #     else:
    #         p = Process(target=_calc_hit_rate_multiprocess_ts,
    #                     args=(reuse_dist_python_list, cache_size, xy_list[divided_len*i:len(xy_list)], q))
    #         process_list.append(p)
    #         p.start()
    #
    # num_left = len(xy_list)
    # while (num_left):
    #     t = q.get()
    #     result[t[0]][t[1]] = t[2]
    #     num_left -= 1
    #     print(num_left)


    # print(result)
    with open('temp', 'wb') as ofile:
        pickle.dump(result, ofile)


def _get_xy_distribution(bin_size, total_length):
    for i in range(0, total_length - bin_size + 1, bin_size):
        # if the total length is not multiple of bin_size, discard the last partition
        for j in range(i + bin_size, total_length - bin_size + 1, bin_size):
            yield (i, j)


def _get_xy_distribution_list(bin_size, total_length):
    l = []
    for i in range(0, total_length - bin_size + 1, bin_size):
        # if the total length is not multiple of bin_size, discard the last partition
        for j in range(i + bin_size, total_length - bin_size + 1, bin_size):
            l.append((i, j))
    return l


def _get_xy_distribution_list_timestamp(reader, time_interval):
    xy_list = []
    xy_dict = dict()
    # generate break point
    # reader.reset()
    break_points = []
    prev = 0
    num = 0
    line = None
    for num, line in enumerate(reader.lines()):
        if (line[0] - prev) > time_interval:
            break_points.append(num)
            prev = line[0]
    # print(num)
    if line[0] != prev:
        break_points.append(num)
    # print(break_points)
    # print(len(break_points))

    if (len(break_points)) > 10000:
        print("number of bins more than 10000, exact size: %d, it may be too slow" % len(break_points))

    return break_points

    # old
    # generate the list for x,y coordinate
    # for i in range(len(break_points)):
    #     # if the total length is not multiple of bin_size, discard the last partition
    #     for j in range(i+1, len(break_points)):
    #         xy_list.append((i, j, break_points[i], break_points[j]))
    #         xy_dict[(i, j)] = (break_points[i], break_points[j])
    #
    # return (xy_list, xy_dict, len(break_points))


def _calc_hit_rate(reuse_dist_array, cache_size, begin_pos, end_pos):
    hit_count = 0
    miss_count = 0
    for i in range(begin_pos, end_pos):
        if reuse_dist_array[i] == -1:
            # never appear
            miss_count += 1
            continue
        if reuse_dist_array[i] - (i - begin_pos) < 0 and reuse_dist_array[i] < cache_size:
            # hit
            hit_count += 1
        else:
            # miss
            miss_count += 1
    # print("hit+miss={}, total size:{}, hit rage:{}".format(hit_count+miss_count, end_pos-begin_pos, hit_count/(end_pos-begin_pos)))
    return hit_count / (end_pos - begin_pos)


def _calc_hit_count(reuse_dist_array, cache_size, begin_pos, end_pos, real_start):
    """

    :rtype: count of hit
    :param reuse_dist_array:
    :param cache_size:
    :param begin_pos:
    :param end_pos:
    :param real_start: the real start position of cache trace
    :return:
    """
    hit_count = 0
    miss_count = 0
    for i in range(begin_pos, end_pos):
        if reuse_dist_array[i] == -1:
            # never appear
            miss_count += 1
            continue
        if reuse_dist_array[i] - (i - real_start) < 0 and reuse_dist_array[i] < cache_size:
            # hit
            hit_count += 1
        else:
            # miss
            miss_count += 1
    # print("hit+miss={}, total size:{}, hit rage:{}".format(hit_count+miss_count, end_pos-begin_pos, hit_count/(end_pos-begin_pos)))
    return hit_count


def _calc_hit_rate_multiprocess(reuse_dist_array, cache_size, xy_list, q):
    for (begin_pos, end_pos) in xy_list:
        hr = _calc_hit_rate(reuse_dist_array, cache_size, begin_pos, end_pos)
        # print('({},{}): {}'.format(begin_pos, end_pos, hr))
        q.put((begin_pos, end_pos, hr))


def _calc_hit_rate_multiprocess_ts(reuse_dist_array, cache_size, xy_list, q):
    for (i, j, begin_pos, end_pos) in xy_list:
        # print(i, j, begin_pos, end_pos)
        hr = _calc_hit_rate(reuse_dist_array, cache_size, begin_pos, end_pos)
        # print('({},{}): {}'.format(i, j, hr))
        q.put((i, j, hr))


def _cal_hit_rate_multiprocess_ts2_init(reuse_dist_array, cache_size, break_points, q):
    _calc_hit_rate_multiprocess_ts2.reuse_dist_array = reuse_dist_array
    _calc_hit_rate_multiprocess_ts2.cache_size = cache_size
    _calc_hit_rate_multiprocess_ts2.break_points = break_points
    _calc_hit_rate_multiprocess_ts2.q = q


def _calc_hit_rate_multiprocess_ts2(order):
    # break_points[i] points to the ith position in reuse_dist_array
    l = []
    total_hc = 0
    for i in range(order + 1, len(_calc_hit_rate_multiprocess_ts2.break_points)):
        hc = _calc_hit_count(_calc_hit_rate_multiprocess_ts2.reuse_dist_array,
                             _calc_hit_rate_multiprocess_ts2.cache_size,
                             _calc_hit_rate_multiprocess_ts2.break_points[i - 1],
                             _calc_hit_rate_multiprocess_ts2.break_points[i],
                             _calc_hit_rate_multiprocess_ts2.break_points[order])
        total_hc += hc
        hr = total_hc / (
            _calc_hit_rate_multiprocess_ts2.break_points[i] - _calc_hit_rate_multiprocess_ts2.break_points[order])
        l.append((order, i, hr))
    # _calc_hit_rate_multiprocess_ts2.q.put(1)
    return l


def draw(filename="heatmap.pdf"):
    with open('temp', 'rb') as ofile:
        result = pickle.load(ofile)
    print("load data successfully, begin plotting")
    # print(result)

    masked_array = np.ma.array(result, mask=np.isnan(result))

    # print(masked_array)
    cmap = plt.cm.jet
    cmap.set_bad('w', 1.)

    plt.Figure()
    heatmap = plt.pcolor(masked_array.T, cmap=cmap)

    # heatmap = plt.pcolor(result2.T, cmap=plt.cm.Blues, vmin=np.min(result2[np.nonzero(result2)]), vmax=result2.max())
    try:
        heatmap = plt.pcolor(masked_array.T, cmap=cmap)
        # heatmap = plt.pcolor(result2.T, cmap=plt.cm.Blues, vmin=np.min(result2[np.nonzero(result2)]), vmax=result2.maxrc())
        plt.show()
        plt.savefig(filename)
    except:
        import matplotlib
        matplotlib.use('pdf')
        heatmap = plt.pcolor(masked_array.T, cmap=cmap)
        plt.savefig(filename)


import os

print(os.getcwd())
# prepare_heatmap_dat(3000, 200)
# prepare_heatmap_dat_multiprocess(1000, 20000, 48)

prepare_heatmap_dat_multiprocess_ts("../data/trace_CloudPhysics_bin", 100000000, 20000, 4, True)
draw()

# for i in range(10, 200, 10):
#     prepare_heatmap_dat_multiprocess_ts("../data/traces/w02_vscsi1.vscsitrace", 1000000000, 200*i*i, 48, True)
#     draw("heatmap_"+str(1000000000)+'_'+str(200*i*i)+'.pdf')



""" mimircache a cache trace analysis platform.

.. moduleauthor:: Juncheng Yang <peter.waynechina@gmail.com>, Ymir Vigfusson

"""

import matplotlib
matplotlib.use('Agg')

import warnings
warnings.filterwarnings("ignore")

# import logging

from mimircache.cache.LRU import LRU as LRU
from mimircache.cacheReader.csvReader import csvCacheReader as csvReader
from mimircache.cacheReader.plainReader import plainCacheReader as plainReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader as vscsiReader
# from mimircache.oldModule.pardaProfiler import pardaProfiler as pardaProfiler
# from mimircache.oldModule.pardaProfiler import parda_mode as parda_mode
from mimircache.profiler.LRUProfiler import LRUProfiler as LRUProfiler
from mimircache.profiler.generalProfiler import generalProfiler as generalProfiler
from mimircache.profiler.heatmap import heatmap as heatmap
from mimircache.top.cachecow import cachecow as cachecow
from mimircache.const import *

from mimircache.const import init

init()


# from _version import __version__

# logging.basicConfig(filename="log", filemode='w', format='%(levelname)s:%(asctime)s:%(message)s', level=logging.DEBUG)

# print(__version__)



from mimircache.cache.LRU import LRU


class clock(LRU):
    """
    second chance page replacement algorithm
    """

    def __init__(self, cache_size=1000):
        # use node id to represent the reference bit
        super(clock, self).__init__(cache_size)
        self.hand = None  # points to the node for examination/eviction

    def _updateElement(self, element):
        """ the given element is in the cache, now update it
        :param element:
        :return: None
        """
        node = self.cacheDict[element]
        node.id = 1

    def _insertElement(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: True on success, False on failure
        """
        if self.cacheLinkedList.size >= self.cache_size:
            self._evictOneElement()

        node = self.cacheLinkedList.insertAtTail(element, id=1)
        self.cacheDict[element] = node
        if not self.hand:
            # this is the first element
            assert self.cacheLinkedList.size == 1, "insert element error"
            self.hand = node

    def _printCacheLine(self):
        for i in self.cacheLinkedList:
            try:
                print("{}({})".format(i.content, i.id), end='\t')
            except:
                print("{}({})".format(i.content, i.id))

        print(' ')

    def _find_evict_node(self):
        node = self.hand
        if node.id == 0:
            self.hand = node.next
            if not self.hand:
                # tail
                self.hand = self.cacheLinkedList.next
            return node
        else:
            # set reference bit to 0
            while node.id == 1:
                node.set_id(0)
                node = node.next
                if not node:
                    # tail
                    node = self.cacheLinkedList.head.next
            self.hand = node.next
            return node

    def _evictOneElement(self):
        """
        evict one element from the cache line
        :return: True on success, False on failure
        """
        node = self._find_evict_node()
        self.cacheLinkedList.removeNode(node)
        del self.cacheDict[node.content]

        return True

    def addElement(self, element):
        '''
        :param element: the element in the reference, it can be in the cache, or not
        :return: None
        '''
        if self.checkElement(element):
            self._updateElement(element)
            # self.printCacheLine()
            if len(self.cacheDict) != self.cacheLinkedList.size:
                print("1*********########### ERROR detected in LRU size #############***********")
                print("{}: {}".format(self.cacheLinkedList.size, len(self.cacheDict)))
                import sys
                sys.exit(-1)
            return True
        else:
            self._insertElement(element)
            # self.printCacheLine()
            if len(self.cacheDict) != self.cacheLinkedList.size:
                print("2*********########### ERROR detected in LRU size #############***********")
                print("{}: {}".format(self.cacheLinkedList.size, len(self.cacheDict)))
                import sys
                sys.exit(-1)
            return False

    def __repr__(self):
        return "second chance cache, given size: {}, current size: {}, {}".format(
            self.cache_size, self.cacheLinkedList.size, super().__repr__())



from mimircache.cache.LRU import LRU


class FIFO(LRU):
    def __init__(self, cache_size=1000):
        super().__init__(cache_size)

    def _updateElement(self, element):
        """ the given element is in the cache, now update it to new location
        :param element:
        :return: None
        """
        pass

    def __repr__(self):
        return "FIFO, given size: {}, current size: {}, {}".format(
            self.cache_size, self.cacheLinkedList.size, super().__repr__())



from mimircache.cache.LRU import LRU
from mimircache.cacheReader.csvReader import csvCacheReader
from mimircache.cacheReader.plainReader import plainCacheReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader
from mimircache.oldModule.basicLRUProfiler import basicLRUProfiler
from mimircache.oldModule.pardaProfiler import pardaProfiler
from mimircache.oldModule.pardaProfiler import parda_mode

# first step: construct a reader for reading any kind of trace

# this one is the most basic one, each line is a label/tag
reader1 = plainCacheReader("../data/parda.trace")

# this one reads csv file and choose one column as label/tag
reader2 = csvCacheReader("../data/trace_CloudPhysics_txt", column=4)

# this one reads binary cloudphysics trace file
reader3 = vscsiCacheReader("../data/trace_CloudPhysics_bin")

# reader is also a generator, for readers you can do the following thing:
# read one trace element at one time:
reader1.read_one_element()
# for loop:
# for element in reader1:
#     pass
# reset, after read some elements, you want to go back
reader1.reset()

# second step: construct a profiler for analyze

# basic mattson profiler (toooooo slow)
basic_profiler = basicLRUProfiler(LRU, cache_size=20000, bin_size=10, reader=reader1)
# or you can use a short way to construct, see below
# basic_profiler = getMRCBasicLRU(LRU, 20000, 10, reader1)

# now let's run it!
# basic_profiler.run()
# # after run, you can either plot it, print it, save the list, save the plot
# basic_profiler.plotHRC()  # the plot is also saved as figure_temp in the data folder in case you forget to save it
# basic_profiler.plotMRC()  # Wow, after run, you can also obtain MRC
# basic_profiler.printHRC()
# basic_profiler.printMRC()
# basic_profiler.outputHRC("I can be a folder")
# basic_profiler.outputMRC("I can also be a file name")



# the second profiler now it supports now is parda
p = pardaProfiler(30000, reader1)  # construction is same, but you don't need to specify bin_size
p.run(parda_mode.seq)  # let's run, it supports two mode, sequential mode and openmp mode
# # p.run(parda_mode.openmp, threads=4)
p.plotHRC()  # the rest is the same as all other profilers, plot, print, output
# p.plotMRC()                         # in case the specified cache-size is not large enough, I also saved
#                                     # parda histogram list in the same folder with data



import sys

RED = '\033[91m'
GREEN = '\033[92m'
YELLOW = '\033[93m'
LIGHT_PURPLE = '\033[94m'
PURPLE = '\033[95m'
BLUE = '\033[34m'
END = '\033[0m'

fcolor_dict = {'red': RED, 'green': GREEN, 'yellow': YELLOW, 'blue': BLUE, 'purple': PURPLE}

bcolor_dict = {'red': '\033[41m', 'green': '\033[42m', 'yellow': '\033[43m', 'blue': '\033[44m', 'purple': '\033[45m',
               'cyan': '\033[46m', 'white': '\033[47m', 'end': '\033[49m'}


def colorfulPrint(color, s):
    print('{}{}{}'.format(fcolor_dict[color.lower()], s, END))


def colorfulPrintWithBackground(bcolor, fcolor, s):
    print('{}{}{}{}'.format(bcolor_dict[bcolor.lower()], fcolor_dict[fcolor.lower()], s, END))


# drop it
def debugPrint(s):
    print("{}: {}: {}".format(sys._getframe().f_code.co_name, sys._getframe().f_lineno, s))


if __name__ == "__main__":
    colorfulPrint("purple", "T")
    colorfulPrintWithBackground("yellow", 'blue', 'text')
    print('normal')
    debugPrint("debug")



import os
import pickle
from collections import deque
from multiprocessing import Array, Process, Queue

import matplotlib
import matplotlib.ticker as ticker
import mimircache.c_heatmap as c_heatmap
import numpy as np
from matplotlib import pyplot as plt

from mimircache.cacheReader.csvReader import csvCacheReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader
from mimircache.profiler.LRUProfiler import LRUProfiler
from mimircache.utils.printing import *
from mimircache.profiler.cHeatmap import cHeatmap


def request_num_2d(reader, mode, time_interval, figname="request_num.png"):
    assert mode == 'r' or mode == 'v', "currently only support mode r and v, what mode are you using?"
    break_points = cHeatmap().gen_breakpoints(reader, mode, time_interval)

    l = []
    for i in range(1, len(break_points)):
        l.append(break_points[i] - break_points[i - 1])
    xticks = ticker.FuncFormatter(lambda x, pos: '{:2.0f}%'.format(x * 100 / len(break_points)))
    draw2d(l, figname=figname, xticks=xticks, xlabel='time({})'.format(mode),
           ylabel='request num count(interval={})'.format(time_interval),
           title='request num count 2D plot')


def cold_miss_2d(reader, mode, time_interval, figname="cold_miss2d.png"):
    assert mode == 'r' or mode == 'v', "currently only support mode r and v, what mode are you using?"
    break_points = cHeatmap().gen_breakpoints(reader, mode, time_interval)

    cold_miss_list = [0] * (len(break_points) - 1)
    seen_set = set()
    never_see = 0
    for i in range(len(break_points) - 1):
        never_see = 0
        for j in range(break_points[i + 1] - break_points[i]):
            r = next(reader)
            if r not in seen_set:
                seen_set.add(r)
                never_see += 1
        cold_miss_list[i] = never_see

    xticks = ticker.FuncFormatter(lambda x, pos: '{:2.0f}%'.format(x * 100 / len(break_points)))
    draw2d(cold_miss_list, figname=figname, xticks=xticks, xlabel='time({})'.format(mode),
           ylabel='cold miss count(interval={})'.format(time_interval),
           title='cold miss count 2D plot')
    reader.reset()


def draw2d(l, **kwargs):
    if 'figname' in kwargs:
        filename = kwargs['figname']
    else:
        filename = '2d_plot.png'

    plt.plot(l)

    if 'xlabel' in kwargs:
        plt.xlabel(kwargs['xlabel'])
    if 'ylabel' in kwargs:
        plt.ylabel(kwargs['ylabel'])
    if 'xticks' in kwargs:
        plt.gca().xaxis.set_major_formatter(kwargs['xticks'])
    if 'yticks' in kwargs:
        plt.gca().yaxis.set_major_formatter(kwargs['yticks'])
    if 'title' in kwargs:
        plt.title(kwargs['title'])

    plt.savefig(filename, dpi=600)
    # plt.show()
    colorfulPrint("red", "plot is saved at the same directory")
    plt.clf()


if __name__ == "__main__":
    pass
    # TIME_INTERVAL = 10000000000
    # from mimircache.cacheReader.vscsiReader import vscsiCacheReader
    # PATH = '/run/shm/traces/'
    #
    # reader = vscsiCacheReader('../data/trace.vscsi')
    # for f in os.listdir(PATH):
    #     if f.endswith('vscsitrace'):
    #         reader = vscsiCacheReader(PATH + f)
    #         request_num_2d(reader, 'r', TIME_INTERVAL,
    #                        figname='0627_request_num/'+f+'_request_num_'+str(TIME_INTERVAL)+'.png')
    # request_num_2d(reader, 'r', 10000000)
    # cold_miss_2d(reader, 'r', 10000000)



class LinkedListNode:
    def __init__(self):
        self.content = None
        self.id = -1
        self.next = None
        self.prev = None

    def set_id(self, id):
        self.id = id


class LinkedList:
    def __init__(self):
        """
        initialization of the linkedlist, head does not contain any element, but head contains last element
        :return:
        """
        self.head = LinkedListNode()
        self.tail = None
        self.size = 0
        self.currentNode = self.head
        # self.max_size = maxlen

    def insertAtTail(self, content, **kargs):
        node = LinkedListNode()
        node.content = content
        if 'id' in kargs:
            node.id = kargs['id']
        node.next = None
        if self.tail:
            # the list is not empty
            node.prev = self.tail
            self.tail.next = node
        else:
            self.head.next = node
            node.prev = self.head
        self.tail = node
        self.size += 1

        # if self.max_size!=-1 and self.size>self.max_size:
        #     self.removeFromTail()

        return node

    def insertNodeAtTail(self, node):
        node.next = None
        if self.tail:
            # the list is not empty
            node.prev = self.tail
            self.tail.next = node
        else:
            self.head.next = node
            node.prev = self.head
        self.tail = node
        self.size += 1

        return node

    def insertAtHead(self, content, **kargs):
        node = LinkedListNode()
        node.content = content
        if 'id' in kargs:
            node.id = kargs['id']
        node.next = self.head.next
        if self.head.next:
            self.head.next.prev = node
        else:
            self.tail = node
        self.head.next = node
        node.prev = self.head
        self.size += 1
        return node

    def removeFromTail(self):
        tail = self.tail
        self.tail.prev.next = None
        temp = self.tail.prev
        self.tail.prev = None
        self.tail = temp
        self.size -= 1
        return tail.content

    def removeFromHead(self):
        if not self.head.next:
            # no node in the list
            raise RuntimeError("there is no element in the list, cannot removeFromHead")
        headContent = self.head.next.content
        temp = self.head.next.next
        if temp:
            # more than one element in the original list
            self.head.next.prev = None
            self.head.next.next = None
            self.head.next = temp
            temp.prev = self.head
        else:
            # there is only one element in the original list, after remove, there will be no element
            del self.head.next
            self.head.next = None
            self.tail = None
        self.size -= 1
        return headContent

    def moveNodeToHead(self, node):
        if self.head.next != node:
            node.prev.next = node.next
            if node.next:
                node.next.prev = node.prev
            else:
                self.tail = node.prev

            node.next = self.head.next
            self.head.next.prev = node
            self.head.next = node
            node.prev = self.head

    def removeNode(self, node):
        node.prev.next = node.next
        if self.tail == node:
            self.tail = node.prev
        else:
            node.next.prev = node.prev
        self.size -= 1
        return node

    def moveNodeToTail(self, node):
        if self.tail != node:
            node.prev.next = node.next
            node.next.prev = node.prev
            node.prev = self.tail
            self.tail.next = node
            node.next = None
            self.tail = node

    def set_node_id(self, node, id):
        node.id = id

    def getHeadContent(self):
        return self.head.next.content

    def getTailContent(self):
        return self.tail.content

    def __iter__(self):
        self.currentNode = self.head
        return self

    def next(self):
        return self.__next__()

    def __next__(self):  # Python 3
        if self.currentNode.next is None:
            # self.currentNode = self.head
            raise StopIteration
        else:
            self.currentNode = self.currentNode.next
            return self.currentNode

    def __repr__(self):
        return "linked list"



"""
.. module:: mimircache.cache
    :platform: Linux, OSX
    :synopsis: The module for all available cache replacement algorithms, currently we have ARC, clock, FIFO,
                LFU_LRU, LFU_MRU, LFU_RR, LRU, MRU, optimal, Random, SLRU, S4LRU,
                To write your own cache replacement algorithm, just use abstractCache.py as base class and override
                all the methods in abstractCache.py. An example is given below:

>>> from mimircache.cache.abstractCache import cache
>>> class Random(cache):
>>>     def __init__(self, cache_size):
>>>         super().__init__(cache_size)
>>>         self.cache_dict = dict()
>>>         self.cache_line_list = []  # to save all the keys, otherwise needs to populate from cache_dict every time
>>>
>>>     def checkElement(self, element):
            '''
            :param element: the key of cache request
            :return: whether the given key is in the cache or not
            '''
>>>         if element in self.cache_dict:
>>>             return True
>>>         else:
>>>             return False
>>>
>>>     def _updateElement(self, element):
            ''' the given element is in the cache, when it is requested again,
             usually we need to update it to new location, but in random, we don't need to do that
            :param element: the key of cache request
            :return: None
            '''
>>>         pass
>>>
>>>     def _insertElement(self, element):
            '''
            the given element is not in the cache, now insert it into cache
            :param element: the key of cache request
            :return: None
            '''
>>>         if len(self.cache_dict) >= self.cache_size:
>>>             self._evictOneElement()
>>>         self.cache_dict[element] = ""
>>>         self.cache_line_list.append(element)
>>>
>>>     def _evictOneElement(self):
            '''
            evict one element from the cache line
            if we delete one element from list every time, it would be O(N) on
            every request, which is too expensive, so we choose to put a hole
            on the list every time we delete it, when there are too many holes
            we re-generate the cache line list
            :return: None
            '''
>>>         rand_num = random.randrange(0, len(self.cache_line_list))
>>>         element = self.cache_line_list[rand_num]
>>>         count = 0
>>>         while not element:
>>>         rand_num = random.randrange(0, len(self.cache_line_list))
>>>         element = self.cache_line_list[rand_num]
>>>         count += 1
            # mark this element as deleted, put a hole on it
>>>         self.cache_line_list[rand_num] = None
>>>         if count > 10:
                # if there are too many holes, then we need to resize the list
>>>             new_list = [e for e in self.cache_line_list if e]
>>>             del self.cache_line_list
>>>             self.cache_line_list = new_list
>>>         del self.cache_dict[element]
>>>
>>>     def addElement(self, element):
            '''
            :param element: the key of cache request, it can be in the cache, or not in the cache
            :return: True if element in the cache
            '''
>>>         if self.checkElement(element):
>>>             self._updateElement(element)
>>>             return True
>>>         else:
>>>             self._insertElement(element)
>>>             return False
>>>
>>>     def __repr__(self):
>>>         return "Random Replacement, given size: {}, current size: {}".format(self.cache_size, len(self.cache_dict), super().__repr__())


.. autoclass:: abstractCache
    :members:

.. autoclass:: LRU

# ARC, clock, FIFO, LFU_LRU, LFU_MRU, LFU_RR, MRU, optimal, RR, S4LRU, SLRU


.. moduleauthor:: Juncheng Yang <peter.waynechina@gmail.com>, Ymir Vigfusson

"""


# __all__ = ['LRU, ARC, clock, FIFO, LFU_LRU, LFU_MRU, LFU_RR, MRU, optimal, RR, S4LRU, SLRU']



import random

from mimircache.cache.abstractCache import cache


class Random(cache):
    def __init__(self, cache_size=1000):
        super().__init__(cache_size)
        self.cacheDict = dict()  # key -> linked list node (in reality, it should also contains value)
        self.cache_line_list = []  # to save all the keys, otherwise needs to populate from cache_dict every time

    def checkElement(self, element):
        """
        :param element: the key of cache request
        :return: whether the given key is in the cache or not
        """
        if element in self.cacheDict:
            return True
        else:
            return False

    def _updateElement(self, element):
        """ the given element is in the cache, when it is requested again,
         usually we need to update it to new location, but in random, we don't need to do that
        :param element: the key of cache request
        :return: None
        """

        pass

    def _insertElement(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element: the key of cache request
        :return: None
        """
        if len(self.cacheDict) >= self.cache_size:
            self._evictOneElement()
        self.cacheDict[element] = ""
        self.cache_line_list.append(element)

    def _printCacheLine(self):
        for i in self.cacheDict:
            try:
                print(i.content, end='\t')
            except:
                print(i.content)

        print(' ')

    def _evictOneElement(self):
        """
        evict one element from the cache line
        if we delete one element from list every time, it would be O(N) on
        every request, which is too expensive, so we choose to put a hole
        on the list every time we delete it, when there are too many holes
        we re-generate the cache line list
        :return: None
        """
        rand_num = random.randrange(0, len(self.cache_line_list))
        element = self.cache_line_list[rand_num]
        count = 0
        while not element:
            rand_num = random.randrange(0, len(self.cache_line_list))
            element = self.cache_line_list[rand_num]
            count += 1

        # mark this element as deleted, put a hole on it
        self.cache_line_list[rand_num] = None

        if count > 10:
            # if there are too many holes, then we need to resize the list
            new_list = [e for e in self.cache_line_list if e]
            del self.cache_line_list
            self.cache_line_list = new_list

        del self.cacheDict[element]

    def addElement(self, element):
        """
        :param element: the key of cache request, it can be in the cache, or not in the cache
        :return: True if element in the cache
        """
        if self.checkElement(element):
            self._updateElement(element)
            return True
        else:
            self._insertElement(element)
            return False

    def __repr__(self):
        return "Random Replacement, given size: {}, current size: {}".format(self.cache_size,
                                                                             len(self.cacheDict),
                                                                             super().__repr__())



from mimircache.cache.LRU import LRU
from mimircache.cache.abstractCache import cache


class S4LRU(cache):
    def __init__(self, cache_size=1000):
        """
        add the fourth part first, then gradually goes up to third, second and first level,
        final eviction is from fourth part
        :param cache_size: size of cache
        :return:
        """
        super(S4LRU, self).__init__(cache_size)

        # Maybe use four linkedlist and a dict will be more efficient?
        self.firstLRU = LRU(self.cache_size // 4)
        self.secondLRU = LRU(self.cache_size // 4)
        self.thirdLRU = LRU(self.cache_size // 4)
        self.fourthLRU = LRU(self.cache_size // 4)

    def checkElement(self, element):
        """
        :param element:
        :return: whether the given element is in the cache
        """
        if element in self.firstLRU or element in self.secondLRU or \
                        element in self.thirdLRU or element in self.fourthLRU:
            return True
        else:
            return False

    def _updateElement(self, element):
        """ the given element is in the cache, now update it to new location
        :param element:
        :return: None
        """
        if element in self.firstLRU:
            self.firstLRU._updateElement(element)
        elif element in self.secondLRU:
            # element is in second, remove from second, insert to end of first,
            # evict from first to second if needed
            self._move_to_upper_level(element, self.secondLRU, self.firstLRU)
        elif element in self.thirdLRU:
            self._move_to_upper_level(element, self.thirdLRU, self.secondLRU)
        elif element in self.fourthLRU:
            self._move_to_upper_level(element, self.fourthLRU, self.thirdLRU)

    def _move_to_upper_level(self, element, lowerLRU, upperLRU):
        """
        move the element from lowerLRU to upperLRU, remove the element from lowerLRU,
        insert into upperLRU, if upperLRU is full, evict one into lowerLRU
        :param element: move the element from lowerLRU to upperLRU
        :param lowerLRU: element in lowerLRU is easier to be evicted
        :param upperLRU: element in higherLRU is evicted into lowerLRU first
        :return:
        """

        # get the node and remove from lowerLRU
        node = lowerLRU.cacheDict[element]
        lowerLRU.cacheLinkedList.removeNode(node)
        del lowerLRU.cacheDict[element]

        # insert into upperLRU
        evicted_key = upperLRU._insertElement(node.content)

        # if there are element evicted from upperLRU, add to lowerLRU
        if evicted_key:
            lowerLRU._insertElement(evicted_key)

    def _insertElement(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: evicted element
        """
        return self.fourthLRU._insertElement(element)

    def _printCacheLine(self):
        print("first: ")
        self.firstLRU._printCacheLine()
        print("second: ")
        self.secondLRU._printCacheLine()
        print("third: ")
        self.thirdLRU._printCacheLine()
        print("fourth: ")
        self.fourthLRU._printCacheLine()

    def _evictOneElement(self):
        """
        evict one element from the cache line
        :return: True on success, False on failure
        """
        pass

    def addElement(self, element):
        """
        :param element: a cache request, it can be in the cache, or not
        :return: None
        """
        if self.checkElement(element):
            self._updateElement(element)
            return True
        else:
            self._insertElement(element)
            return False

    def __repr__(self):
        return "S4LRU, given size: {}, current 1st part size: {}, current 2nd size: {}, \
            current 3rd part size: {}, current fourth part size: {}". \
            format(self.cache_size, self.firstLRU.cacheLinkedList.size, self.secondLRU.cacheLinkedList.size,
                   self.thirdLRU.cacheLinkedList.size, self.fourthLRU.cacheLinkedList.size)



from mimircache.cache.abstractCache import cache
from mimircache.utils.LinkedList import LinkedList


class LRU(cache):
    def __init__(self, cache_size=1000):
        super().__init__(cache_size)
        self.cacheLinkedList = LinkedList()
        self.cacheDict = dict()  # key -> linked list node (in reality, it should also contains value)

    def checkElement(self, element):
        """
        :param element:
        :return: whether the given element is in the cache
        """
        if element in self.cacheDict:
            return True
        else:
            return False

    def _updateElement(self, element):
        """ the given element is in the cache, now update it to new location
        :param element:
        :return: None
        """

        node = self.cacheDict[element]
        self.cacheLinkedList.moveNodeToTail(node)

    def _insertElement(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: evicted element or None
        """
        return_content = None
        if self.cacheLinkedList.size >= self.cache_size:
            # print("{}: {}".format(self.cacheLinkedList.size, self.cache_size))
            # if self.cacheLinkedList.size == self.cache_size:
            #     self.printCacheLine()
            # print(self.cache_dict)
            return_content = self._evictOneElement()

        node = self.cacheLinkedList.insertAtTail(element)
        self.cacheDict[element] = node
        return return_content

    def _printCacheLine(self):
        for i in self.cacheLinkedList:
            try:
                print(i.content, end='\t')
            except:
                print(i.content)

        print(' ')

    def _evictOneElement(self):
        """
        evict one element from the cache line
        :return: content of evicted element
        """
        content = self.cacheLinkedList.removeFromHead()
        # print(content)
        # self.printCacheLine()
        # print(self.cache_dict)
        del self.cacheDict[content]
        return content

    def addElement(self, element):
        """
        :param element: the element in the reference, it can be in the cache, or not
        :return: None
        """
        if self.checkElement(element):
            self._updateElement(element)
            if len(self.cacheDict) != self.cacheLinkedList.size:
                print("1*********########### ERROR detected in LRU size #############***********")
                print("{}: {}".format(self.cacheLinkedList.size, len(self.cacheDict)))
                import sys
                sys.exit(-1)
            return True
        else:
            self._insertElement(element)
            if len(self.cacheDict) != self.cacheLinkedList.size:
                print("2*********########### ERROR detected in LRU size #############***********")
                print("{}: {}".format(self.cacheLinkedList.size, len(self.cacheDict)))
                import sys
                sys.exit(-1)
            return False

    def __repr__(self):
        return "LRU, given size: {}, current size: {}, {}".format(self.cache_size, self.cacheLinkedList.size,
                                                                  super().__repr__())



""" this module is used for all other cache replacement algorithms excluding LRU(LRU also works, but slow compared to
    using pardaProfiler),
"""
# -*- coding: utf-8 -*-


import math
import os
import time

import traceback

import numpy as np
from multiprocessing import Process, Pipe, Array
from os import path

from mimircache.cacheReader.vscsiReader import vscsiCacheReader

from mimircache.cache.adaptiveSLRU import AdaptiveSLRU
from mimircache.cache.clock import clock
from mimircache.cache.FIFO import FIFO
from mimircache.cache.LFU_LRU__NEED_OPTIMIZATION import LFU_LRU
from mimircache.cache.LFU_MRU import LFU_MRU
from mimircache.cache.LFU_RR import LFU_RR
from mimircache.cache.LRU import LRU
from mimircache.cache.MRU import MRU
from mimircache.cache.Random import Random
from mimircache.cache.SLRU import SLRU
from mimircache.cache.S4LRU import S4LRU
from mimircache.cache.Optimal import optimal


from mimircache.cacheReader.plainReader import plainCacheReader

import mimircache.c_generalProfiler as c_generalProfiler

import matplotlib.pyplot as plt

from mimircache.profiler.abstract.abstractProfiler import profilerAbstract

from mimircache.const import *


class generalProfiler(profilerAbstract):
    def __init__(self, reader, cache_class, cache_size, bin_size=-1, cache_params=None,
                 num_of_process=DEFAULT_NUM_OF_PROCESS):
        super(generalProfiler, self).__init__(cache_class, cache_size, reader)
        self.cache_params = cache_params
        self.num_of_process = num_of_process

        if bin_size == -1:
            self.bin_size = int(self.cache_size / DEFAULT_BIN_NUM_PROFILER)
        else:
            self.bin_size = bin_size
        self.num_of_process = num_of_process

        if self.cache_size != -1:

            self.num_of_cache = self.num_of_blocks = math.ceil(self.cache_size / bin_size)

            self.HRC = np.zeros((self.num_of_blocks + 1,), dtype=np.double)
            self.MRC = np.zeros((self.num_of_blocks + 1,), dtype=np.double)

        else:
            raise RuntimeError("you input -1 as cache size")
        self.cache = None
        self.cache_list = None

        self.process_list = []
        self.cache_distribution = [[] for _ in range(self.num_of_process)]

        # shared memory for storing MRC count
        self.MRC_shared_array = Array('i', range(self.num_of_blocks))
        for i in range(len(self.MRC_shared_array)):
            self.MRC_shared_array[i] = 0

        # dispatch different cache size to different processes, does not include size = 0
        for i in range(self.num_of_blocks):
            self.cache_distribution[i % self.num_of_process].append((i + 1) * bin_size)

            # build pipes for communication between main process and children process
            # the pipe mainly sends element from main process to children
        self.pipe_list = []
        for i in range(self.num_of_process):
            self.pipe_list.append(Pipe())
            p = Process(target=self._addOneTraceElementSingleProcess,
                        args=(self.num_of_process, i, self.cache_class, self.cache_distribution[i],
                              self.cache_params, self.pipe_list[i][1], self.MRC_shared_array))
            self.process_list.append(p)
            p.start()

        self.calculated = False

    all = ["get_hit_count", "get_hit_rate", "get_miss_rate", "plotMRC", "plotHRC"]


    def addOneTraceElement(self, element):
        super().addOneTraceElement(element)

        for i in range(len(self.pipe_list)):
            # print("send out: " + element)
            self.pipe_list[i][0].send(element)

        return

    # noinspection PyMethodMayBeStatic
    def _addOneTraceElementSingleProcess(self, num_of_process, process_num, cache_class, cache_size_list,
                                         cache_args, pipe, MRC_shared_array):
        """

        :param num_of_process:
        :param process_num:
        :param cache_class:
        :param cache_size_list: a list of different cache size dispached to this process
        :param cache_args: the extra argument (besides size) for instantiate a cache class
        :param pipe:            for receiving cache record from main process
        :param MRC_shared_array:       storing MRC count for all cache sizes
        :return:
        """
        cache_list = []
        for i in range(len(cache_size_list)):
            if cache_args:
                cache_list.append(cache_class(cache_size_list[i], **cache_args))
            else:
                cache_list.append(cache_class(cache_size_list[i]))

        elements = pipe.recv()

        # TODO this part should be changed
        while elements[-1] != 'END_1a1a11a_ENDMARKER':
            for i in range(len(cache_list)):
                # print("i = %d"%i)
                # cache_list[i].printCacheLine()
                # print('')
                for element in elements:
                    if not cache_list[i].addElement(element):
                        MRC_shared_array[i * num_of_process + process_num] += 1
            elements = pipe.recv()
            # print(element)
            # print(cache_list)


    def run(self, buffer_size=10000):
        super().run()
        self.reader.reset()
        l = []
        for i in self.reader:
            l.append(i)
            if len(l) == buffer_size:
                self.add_elements(l)
                l.clear()
                # self.addOneTraceElement(i)
        # p.printMRC()
        if len(l) > 0:
            self.add_elements(l)
        self.calculate()
        # self.outputHRC()
        # self.plotHRC()

    def add_elements(self, elements):
        for element in elements:
            super().addOneTraceElement(element)

        for i in range(len(self.pipe_list)):
            # print("send out: " + element)
            self.pipe_list[i][0].send(elements)

        return

    def calculate(self):
        self.calculated = True
        for i in range(len(self.pipe_list)):
            self.pipe_list[i][0].send(["END_1a1a11a_ENDMARKER"])
            self.pipe_list[i][0].close()
        for i in range(len(self.process_list)):
            self.process_list[i].join()

        self.MRC[0] = 1
        for i in range(1, len(self.MRC), 1):
            # print(self.MRC_shared_array[i])
            self.MRC[i] = self.MRC_shared_array[i - 1] / self.num_of_trace_elements
        for i in range(1, len(self.HRC), 1):
            self.HRC[i] = 1 - self.MRC[i]

    def get_hit_count(self):
        if not self.calculated:
            self.run()

        HC = np.zeros((self.num_of_blocks + 1,), dtype=np.longlong)
        for i in range(1, len(HC), 1):
            HC[i] = self.num_of_trace_elements - self.MRC_shared_array[i - 1]

        return HC

    def get_hit_rate(self):
        if not self.calculated:
            self.run()
        return self.HRC

    def get_miss_rate(self):
        if not self.calculated:
            self.run()
        return self.MRC

    def plotMRC(self, figname="MRC.png", **kwargs):
        if not self.calculated:
            self.calculate()
        try:
            num_of_blocks = self.num_of_blocks + 1
            plt.plot(range(0, self.bin_size * num_of_blocks, self.bin_size), self.MRC[:num_of_blocks])
            plt.xlabel("cache Size")
            plt.ylabel("Miss Rate")
            plt.title('Miss Rate Curve', fontsize=18, color='black')
            plt.savefig(figname)
            plt.show()
            plt.clf()
        except Exception as e:
            plt.savefig(figname)
            print("the plotting function is not wrong, is this a headless server?")
            print(e)
            traceback.print_exc()

    def plotHRC(self, figname="HRC.png", **kwargs):
        if not self.calculated:
            self.calculate()
        try:
            num_of_blocks = self.num_of_blocks + 1

            plt.plot(range(0, self.bin_size * num_of_blocks, self.bin_size), self.HRC[:num_of_blocks])
            plt.xlabel("cache Size")
            plt.ylabel("Hit Rate")
            plt.title('Hit Rate Curve', fontsize=18, color='black')
            plt.savefig(figname)
            plt.show()
            plt.clf()
        except Exception as e:
            plt.savefig(figname)
            print("the plotting function is not wrong, is this a headless server?")
            print(e)




if __name__ == "__main__":
    import time

    t1 = time.time()
    # r = plainCacheReader('../data/test.dat')
    # r = plainCacheReader('../data/parda.trace')
    r = vscsiCacheReader('../data/trace.vscsi')

    arc_dict = {'p': 0.5, 'ghostlist_size': -1}
    # p = generalProfiler(r, ARC, 1000, 100, arc_dict, 8)

    p = generalProfiler(r, Random, 3000, 200, num_of_process=8)
    print(p.get_hit_rate())
    print(p.get_hit_count())
    t2 = time.time()
    print("TIME: %f" % (t2 - t1))

    t1 = time.time()
    p.plotMRC()
    # hr = c_generalProfiler.get_hit_rate(r.cReader, 2000, "Optimal", bin_size=200, num_of_threads=8)
    # print(hr)

    t2 = time.time()
    print("TIME: %f" % (t2 - t1))

    # for i in r:
    #     # print(i)
    #     p.addOneTraceElement(i)
    # # p.printMRC()
    # p.plotHRC()



import logging
import os

import matplotlib
import matplotlib.ticker as ticker
import mimircache.c_heatmap as c_heatmap
import numpy as np
from matplotlib import pyplot as plt

from mimircache import const
from mimircache.cacheReader.vscsiReader import vscsiCacheReader
from mimircache.utils.printing import *


class cHeatmap:
    def __init__(self):
        self.other_plot_kwargs = {}

    def gen_breakpoints(self, reader, mode, time_interval):
        """

        :param reader:
        :param mode:
        :param time_interval:
        :return: a numpy list of break points begin with 0, ends with total_num_requests
        """
        return c_heatmap.gen_breakpoints(reader.cReader, mode, time_interval)

    def set_plot_params(self, axis, axis_type, **kwargs):
        log_base = 1
        label = ''
        tick = None
        xydict = None

        if 'xydict' in kwargs:
            xydict = kwargs['xydict']

        if 'log_base' in kwargs:
            log_base = kwargs['log_base']

        if 'fixed_range' in kwargs and kwargs['fixed_range']:
            # if 'vmin' in kwargs:
            #     self.other_plot_kwargs['vmin'] = kwargs['vmin']
            # vmin, vmax = kwargs['fixed_range']
            # self.other_plot_kwargs['vmin'] = 0
            # self.other_plot_kwargs['vmax'] = 1
            self.other_plot_kwargs['fixed_range'] = kwargs['fixed_range']

        if 'other_kwargs' in kwargs:
            self.other_plot_kwargs.update(kwargs['other_kwargs'])

        if 'text' in kwargs:
            ax = plt.gca()
            ax.text(kwargs['text'][0], kwargs['text'][1], kwargs['text'][2])  # , fontsize=20)  , color='blue')

        if axis == 'x':
            assert 'xydict' in kwargs, "you didn't provide xydict"
            array = xydict[0]
        elif axis == 'y':
            assert 'xydict' in kwargs, "you didn't provide xydict"
            array = xydict
        elif axis == 'cb':  # color bar
            if axis_type == 'count':
                self.other_plot_kwargs['norm'] = matplotlib.colors.LogNorm()
        else:
            print("unsupported axis: " + str(axis))

        if axis_type == 'virtual_time':
            if 'label' in kwargs:
                label = kwargs['label']
            else:
                label = 'virtual time'
            # tick = ticker.FuncFormatter(lambda x, pos: '{:2.0f}'.format(bin_size * x))
            tick = ticker.FuncFormatter(lambda x, pos: '{:2.0f}%'.format(x * 100 / (len(array) - 1)))
        elif axis_type == 'real_time':
            if 'label' in kwargs:
                label = kwargs['label']
            else:
                label = 'real time'
            # tick = ticker.FuncFormatter(lambda x, pos: '{:2.0f}'.format(bin_size * x))
            tick = ticker.FuncFormatter(lambda x, pos: '{:2.0f}%'.format(x * 100 / (len(array) - 1)))

        elif axis_type == 'cache_size':
            assert log_base != 1, "please provide log_base"
            label = 'cache size'
            tick = ticker.FuncFormatter(lambda x, pos: '{:2.0f}'.format(log_base ** x))

        elif axis_type == 'reuse_dist':
            assert log_base != 1, "please provide log_base"
            label = 'reuse distance'
            tick = ticker.FuncFormatter(lambda x, pos: '{:2.0f}'.format(log_base ** x))

        if axis == 'x':
            plt.xlabel(label)
            plt.gca().xaxis.set_major_formatter(tick)

        elif axis == 'y':
            plt.ylabel(label)
            plt.gca().yaxis.set_major_formatter(tick)

        elif axis == 'cb':
            pass

        else:
            print("unsupported axis: " + str(axis))

    def heatmap(self, reader, mode, time_interval, plot_type, algorithm="LRU", cache_params=None, **kwargs):
        """

        :param plot_type:
        :param mode:
        :param interval:
        :param reader:
        :param kwargs: include num_of_process, figname
        :return:
        """

        figname = None
        self.time_interval = time_interval
        self.mode = mode
        reader.reset()

        if 'cache_size' in kwargs:
            cache_size = kwargs['cache_size']
        else:
            cache_size = -1            

        if 'num_of_threads' in kwargs:
            num_of_threads = kwargs['num_of_threads']
        else:
            num_of_threads = 4

        if 'figname' in kwargs:
            figname = kwargs['figname']

        if mode == 'r' or mode == 'v':
            if plot_type == "hit_rate_start_time_end_time":
                # assert algorithm!=None, "please specify your cache replacement algorithm in heatmap plotting"
                assert cache_size != -1, "please provide cache_size parameter for plotting hit_rate_start_time_end_time"
                print("going to plot heatmap of hit_rate_start_time_end_time using {} "
                      "of cache size: {}".format(algorithm, cache_size))

                if algorithm.lower() in const.c_available_cache:
                    print('time: {}, size: {}, threads: {}'.format(time_interval, cache_size, num_of_threads))
                    xydict = c_heatmap.heatmap(reader.cReader, mode, time_interval, plot_type,
                                               cache_size, algorithm, cache_params=cache_params,
                                               num_of_threads=num_of_threads)
                else:
                    raise RuntimeError("haven't provide support given algorithm yet: " + str(algorithm))
                    pass

                text = "      " \
                       "cache size: {},\n      cache type: {},\n      " \
                       "time type:  {},\n      time interval: {},\n      " \
                       "plot type: \n{}".format(cache_size,
                                                algorithm, mode, time_interval, plot_type)

                x1, y1 = xydict.shape
                x1 = int(x1 / 2.8)
                y1 /= 8
                if mode == 'r':
                    self.set_plot_params('x', 'real_time', xydict=xydict, label='start time (real)',
                                         text=(x1, y1, text))
                    self.set_plot_params('y', 'real_time', xydict=xydict, label='end time (real)', fixed_range=(0, 1))
                else:
                    self.set_plot_params('x', 'virtual_time', xydict=xydict, label='start time (virtual)',
                                         text=(x1, y1, text))
                    self.set_plot_params('y', 'virtual_time', xydict=xydict, label='end time (virtual)',
                                         fixed_range=(0, 1))

                if not figname:
                    figname = '_'.join([algorithm, str(cache_size), plot_type]) + '.png'
                self.draw_heatmap(xydict, figname=figname)



            elif plot_type == "hit_rate_start_time_cache_size":
                pass


            elif plot_type == "avg_rd_start_time_end_time":
                pass


            elif plot_type == "cold_miss_count_start_time_end_time":
                print("this plot is discontinued")


            elif plot_type == "???":
                pass




            elif plot_type == "rd_distribution":
                if not figname:
                    figname = 'rd_distribution.png'

                xydict, log_base = c_heatmap.heatmap_rd_distribution(reader.cReader, mode, time_interval,
                                                                     num_of_threads=num_of_threads)
                self.set_plot_params('x', 'real_time', xydict=xydict)
                self.set_plot_params('y', 'reuse_dist', xydict=xydict, log_base=log_base)
                self.set_plot_params('cb', 'count')
                self.draw_heatmap(xydict, figname=figname, not_mask=True)

            elif plot_type == "rd_distribution_CDF":
                if not figname:
                    figname = 'rd_distribution_CDF.png'

                xydict, log_base = c_heatmap.heatmap_rd_distribution(reader.cReader, mode, time_interval,
                                                                     num_of_threads=num_of_threads, CDF=1)
                self.set_plot_params('x', 'real_time', xydict=xydict)
                self.set_plot_params('y', 'reuse_dist', xydict=xydict, log_base=log_base)
                # self.set_plot_params('cb', 'count')
                self.draw_heatmap(xydict, figname=figname, not_mask=True)


            elif plot_type == "future_rd_distribution":
                if not figname:
                    figname = 'future_rd_distribution.png'

                xydict, log_base = c_heatmap.heatmap_future_rd_distribution(reader.cReader, mode, time_interval,
                                                                            num_of_threads=num_of_threads)
                self.set_plot_params('x', 'real_time', xydict=xydict)
                self.set_plot_params('y', 'reuse_dist', xydict=xydict, log_base=log_base)
                self.set_plot_params('cb', 'count')
                self.draw_heatmap(xydict, figname=figname, not_mask=True)


            else:
                raise RuntimeError(
                    "Cannot recognize this plot type, please check documentation, yoru input is %s" % mode)

        else:
            raise RuntimeError("Cannot recognize this mode, it can only be either real time(r) or virtual time(v), "
                               "but you input %s" % mode)

        reader.reset()

    def differential_heatmap(self, reader, mode, time_interval, plot_type, algorithm1, algorithm2="Optimal",
                             cache_params1=None, cache_params2=None, **kwargs):
        """

        :param plot_type:
        :param mode:
        :param interval:
        :param reader:
        :param kwargs: include num_of_process, figname
        :return:
        """

        self.time_interval = time_interval
        self.mode = mode
        reader.reset()

        if 'cache_size' in kwargs:
            cache_size = kwargs['cache_size']
        else:
            cache_size = -1

        figname = '_'.join([algorithm2 + '-' + algorithm1, str(cache_size), plot_type]) + '.png'

        if 'num_of_threads' in kwargs:
            num_of_threads = kwargs['num_of_threads']
        else:
            num_of_threads = 4

        if 'figname' in kwargs:
            figname = kwargs['figname']

        if mode == 'r' or mode == 'v':
            if plot_type == "hit_rate_start_time_end_time":
                assert cache_size != -1, "please provide cache_size for plotting hit_rate_start_time_end_time"
                print("going to plot differential heatmap of hit_rate_start_time_end_time using {} - {} "
                      "of cache size: {}".format(algorithm2, algorithm1, cache_size))

                xydict = c_heatmap.differential_heatmap(reader.cReader, mode, time_interval,
                                                        plot_type, cache_size, algorithm1, algorithm2,
                                                        cache_params1=cache_params1, cache_params2=cache_params2,
                                                        num_of_threads=num_of_threads)
                # xydict = c_heatmap.differential_heatmap_with_Optimal(reader.cReader, cache_size, alg1, mode,
                #                                         time_interval,
                #                                         plot_type, num_of_threads=num_of_threads)

                text = "      differential heatmap\n      cache size: {},\n      cache type: {}-{},\n" \
                       "      time type: {},\n      time interval: {},\n      plot type: \n{}".format(
                    cache_size, algorithm2, algorithm1, mode, time_interval, plot_type)

                x1, y1 = xydict.shape
                x1 = int(x1 / 2.8)
                y1 /= 8
                if mode == 'r':
                    self.set_plot_params('x', 'real_time', xydict=xydict, label='start time (real)',
                                         text=(x1, y1, text))
                    self.set_plot_params('y', 'real_time', xydict=xydict, label='end time (real)', fixed_range=(-1, 1))
                else:
                    self.set_plot_params('x', 'virtual_time', xydict=xydict, label='start time (virtual)',
                                         text=(x1, y1, text))
                    self.set_plot_params('y', 'virtual_time', xydict=xydict, label='end time (virtual)',
                                         fixed_range=(-1, 1))

                self.draw_heatmap(xydict, figname=figname)



            elif plot_type == "hit_rate_start_time_cache_size":
                pass


            elif plot_type == "avg_rd_start_time_end_time":
                pass


            elif plot_type == "cold_miss_count_start_time_end_time":
                print("this plot is discontinued")


            elif plot_type == "???":
                pass



            else:
                raise RuntimeError(
                    "Cannot recognize this plot type, please check documentation, yoru input is %s" % mode)

        else:
            raise RuntimeError("Cannot recognize this mode, it can only be either real time(r) or virtual time(v), "
                               "but you input %s" % mode)

        reader.reset()

    def draw_heatmap(self, xydict, **kwargs):
        if 'figname' in kwargs:
            filename = kwargs['figname']
        else:
            filename = 'heatmap.png'

        if 'not_mask' in kwargs and kwargs['not_mask']:
            plot_array = xydict
        else:
            plot_array = np.ma.array(xydict, mask=np.tri(len(xydict), dtype=int).T)

        cmap = plt.cm.jet
        cmap.set_bad('w', 1.)

        # plt.yscale('log')

        plt.title("Heatmap")

        try:
            # ax.pcolor(masked_array.T, cmap=cmap)
            # print(str(**self.other_plot_kwargs))
            if 'fixed_range' in self.other_plot_kwargs and self.other_plot_kwargs['fixed_range']:
                vmin, vmax = self.other_plot_kwargs['fixed_range']
                del self.other_plot_kwargs['fixed_range']
                img = plt.imshow(plot_array, vmin=vmin, vmax=vmax, interpolation='nearest', origin='lower',
                                 aspect='auto', **self.other_plot_kwargs)
            else:
                img = plt.imshow(plot_array, interpolation='nearest', origin='lower', aspect='auto',
                                 **self.other_plot_kwargs)

            # , vmin=0, vmax=1

            cb = plt.colorbar(img)

            plt.savefig(filename, dpi=600)
            plt.show()
            colorfulPrint("red", "plot is saved at the same directory")
            plt.clf()
            self.other_plot_kwargs.clear()
        except Exception as e:
            logging.warning(str(e))
            plt.pcolormesh(plot_array.T, cmap=cmap)
            plt.savefig(filename)


def server_plot_all(path="../data/cloudphysics/", num_of_threads=48):
    TIME_INTERVAL = 1000000000
    PLOT_TYPE = "hit_rate_start_time_end_time"
    MODE = 'r'
    mem_sizes = [int(200 * (i ** 2)) for i in range(1, 20)]
    # with open('memSize', 'r') as ifile:
    #     for line in ifile:
    #         mem_sizes.append(int(line.strip()))
    # mem_sizes.append()

    for filename in os.listdir(path):
        print(filename)
        if filename.endswith('.vscsitrace'):
            hm = cHeatmap()
            # mem_size = mem_sizes[int(filename.split('_')[0][1:]) - 1] * 16
            reader = vscsiCacheReader(path + filename)
            # TIME_INTERVAL = int(reader.get_num_of_total_requests() / 200)
            # mem_sizes = LRUProfiler(reader).get_best_cache_sizes(12, 200, 20)

            # if os.path.exists("0624/" + '_'.join([filename, "_LRU_Optimal_", str(mem_sizes[-1]), MODE]) + "_.png"):
            #     continue

            # t1 = time.time()
            # hm.heatmap(reader, MODE, TIME_INTERVAL, "rd_distribution_CDF", num_of_threads=num_of_threads, figname=
            # "0624/" + '_'.join([filename, "_rd_distribution_CDF", MODE]) + "_.png")
            # print("time: " + str(time.time() - t1))

            # t1 = time.time()
            # hm.heatmap(reader, MODE, TIME_INTERVAL, "future_rd_distribution", num_of_threads=num_of_threads,
            #            figname="0620/" + '_'.join([filename, "_future_rd_distribution_", MODE]) + "_.png")
            # print("time: " + str(time.time() - t1))


            for mem_size in mem_sizes:
                t1 = time.time()
                hm.differential_heatmap(reader, MODE, TIME_INTERVAL, PLOT_TYPE, "LRU", "LRU_K",
                                        cache_size=mem_size, cache_params2={"K": 2},
                                        num_of_threads=num_of_threads,
                                        figname="0627_LRUK_comp/" + '_'.join(
                                            [filename, "_LRU_LRU2_", str(mem_size), MODE]) + "_.png")
                print("cache size: " + str(mem_size) + ", time: " + str(time.time() - t1))

                t1 = time.time()
                hm.differential_heatmap(reader, MODE, TIME_INTERVAL, PLOT_TYPE, "LRU", "LRU_K",
                                        cache_size=mem_size, cache_params2={"K": 3},
                                        num_of_threads=num_of_threads,
                                        figname="0627_LRUK_comp/" + '_'.join(
                                            [filename, "_LRU_LRU3_", str(mem_size), MODE]) + "_.png")
                print("cache size: " + str(mem_size) + ", time: " + str(time.time() - t1))

                t1 = time.time()
                hm.differential_heatmap(reader, MODE, TIME_INTERVAL, PLOT_TYPE, "LRU", "LRU_K",
                                        cache_size=mem_size, cache_params2={"K": 4},
                                        num_of_threads=num_of_threads,
                                        figname="0627_LRUK_comp/" + '_'.join(
                                            [filename, "_LRU_LRU4_", str(mem_size), MODE]) + "_.png")
                print("cache size: " + str(mem_size) + ", time: " + str(time.time() - t1))

            reader.close()

            del hm
            del reader


def localtest():
    TIME_INTERVAL = 10000000
    PLOT_TYPE = "hit_rate_start_time_end_time"
    MODE = 'r'
    NUM_OF_THREADS = 8
    CACHE_SIZE = 2000

    reader = vscsiCacheReader('../data/trace.vscsi')

    hm = cHeatmap()

    print('a')
    # hm.heatmap(reader, MODE, TIME_INTERVAL, "rd_distribution_CDF", num_of_threads=NUM_OF_THREADS,
    #            figname="rd_dist_CDF_relative.png")

    print('b')
    # hm.heatmap(reader, MODE, TIME_INTERVAL, PLOT_TYPE, cache_size=CACHE_SIZE, num_of_threads=int(NUM_OF_THREADS),
    #            algorithm="LRU_K", cache_params={'K':20},
    #            figname=PLOT_TYPE + ".png")


    print('c')
    t1 = time.time()
    hm.differential_heatmap(reader, MODE, TIME_INTERVAL, PLOT_TYPE, "LRU", "LRU_K", cache_params2={'K': 3},
                            cache_size=CACHE_SIZE,
                            num_of_threads=NUM_OF_THREADS, figname="heatmap.png")
    print(time.time() - t1)


def checker(xydict):
    for i in range(len(xydict)):
        for j in range(len(xydict[0])):
            if xydict[i][j] > 1 or xydict[i][j] < 0:
                print("{},{}: {}".format(i, j, xydict[i][j]))


if __name__ == "__main__":
    import time

    t1 = time.time()
    # server_plot_all(path="/run/shm/traces/")
    # server_plot_all(path="../../../../disk/traces/", num_of_threads=16)
    localtest()

    CACHESIZE = 327680
    BINSIZE = 1
    TIME_INTERVAL = 10000000
    CACHE_TYPE = "Optimal"
    TIME_TYPE = 'r'
    PLOT_TYPE = "hit_rate_start_time_end_time"
    NUM_OF_THREADS = 8

    from mimircache.profiler.heatmap import heatmap

    hm = cHeatmap()
    reader = vscsiCacheReader('../data/trace.vscsi')
    print(hm.gen_breakpoints(reader, 'r', 1000000))

    # print(heatmap()._get_breakpoints_realtime(reader, 1000000))
    # reader = vscsiCacheReader('/home/cloudphysics/traces/w84_vscsi1.vscsitrace')





import matplotlib.pyplot as plt


class KeyTrace:
    def __init__(self, reader, num_keys):
        self.reader = reader
        self.num_keys = num_keys

    def show(self, num_colors=4, min_frequency=0):
        x, y = self.generate_plot_data(min_frequency=min_frequency)

        colors = []
        color = 0
        colors.append(color)

        for e in range(1, len(y)):
            if y[e - 1] != y[e]:
                color += 1
                if color == num_colors:
                    color = 0

            colors.append(color)

        plt.scatter(x, y, marker='s', c=colors, edgecolors="white", alpha=0.5)
        plt.xlabel("Time")
        plt.ylabel("Key")
        plt.show()

    def generate_plot_data(self, min_frequency=0):
        if min_frequency < 0:
            raise Exception("Invalid min_frequency: value must be positive")

        logical_time = 0
        d = {}

        if min_frequency > 0:
            for element in self.reader:
                if element in d:
                    d[element].append(logical_time)
                else:
                    d[element] = []

                logical_time += 1

            for key, value in list(d.items()):
                if len(value) < min_frequency:
                    del d[key]

            values = sorted(d.values(), key=len)[len(d) - 1 - self.num_keys:len(d) - 1]

        else:
            for element in self.reader:

                if element in d:
                    d[element].append(logical_time)
                elif len(d) < self.num_keys:
                    d[element] = []

                logical_time += 1

            values = d.values()

        x = []
        y = []
        key = 0

        for item in values:
            for time in item:
                x.append(time)
                y.append(key)

            key += 1

        return x, y



"""

"""

# still CPU intensive, needs to find out why and optimize


import random
from mimircache.cache.abstractLFU import abstractLFU


class LFU_RR(abstractLFU):
    def __init__(self, cache_size=1000):
        super(LFU_RR, self).__init__(cache_size)

    def find_evict_key(self):
        r = random.randrange(0, len(self.least_freq_elements_list))
        evict_key = self.least_freq_elements_list[r]
        count = 0
        while not evict_key:
            r = random.randrange(0, len(self.least_freq_elements_list))
            evict_key = self.least_freq_elements_list[r]
            count += 1
        self.least_freq_elements_list[r] = None

        if count > 10:
            new_list = [e for e in self.least_freq_elements_list if e]
            del self.least_freq_elements_list
            self.least_freq_elements_list = new_list

        return evict_key


if __name__ == "__main__":
    pass



import logging
from ctypes import *
from enum import Enum

from mimircache.cache.LRU import LRU
from mimircache.cacheReader.plainReader import plainCacheReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader
from mimircache.profiler.abstract.abstractLRUProfiler import abstractLRUProfiler


# TODO: to optimize use numpy for ctype return value
# from numpy.ctypeslib import ndpointer


class parda_mode(Enum):
    seq = 1
    openmp = 2
    mpi = 3
    hybrid = 4


class pardaProfiler(abstractLRUProfiler):
    def __init__(self, cache_size, reader, bin_size=1):
        super().__init__(LRU, cache_size, bin_size, reader)

        # load the shared object file
        self.parda_seq = CDLL(os.path.join(os.path.dirname(os.path.abspath(__file__)), self.get_lib_name()))

        self.c_cache_size = c_long(self.cache_size)

        # if the given file is not basic reader, needs conversion
        if not isinstance(reader, plainCacheReader):
            self.prepare_file()
        else:
            self.num_of_trace_elements = self.reader.get_num_of_total_requests()

        self.calculated = False

    def get_lib_name(self):
        for name in os.listdir(os.path.dirname(os.path.abspath(__file__))):
            if 'libparda' in name and '.py' not in name:
                return name

    def prepare_file_remove_one(self):
        """
        this function will prepare the file, meanwhile remove the request that appear only once
        :return:
        """
        self.num_of_trace_elements = 0
        logging.debug("changing file format")
        seen_dict = {}
        for e in self.reader:
            seen_dict[e] = seen_dict.get(e, 0) + 1
        self.reader.reset()
        print(len(seen_dict))
        with open('temp.dat', 'w') as ofile:
            i = self.reader.read_one_element()
            while i is not None:
                self.num_of_trace_elements += 1
                if seen_dict[i] > 1:
                    ofile.write(str(i) + '\n')
                i = self.reader.read_one_element()
        self.reader = plainCacheReader('temp.dat')
        print(self.num_of_trace_elements)

    def prepare_file(self):
        self.num_of_trace_elements = 0
        logging.debug("changing file format")
        with open('temp.dat', 'w') as ofile:
            i = self.reader.read_one_element()
            while i is not None:
                self.num_of_trace_elements += 1
                ofile.write(str(i) + '\n')
                i = self.reader.read_one_element()
        self.reader = plainCacheReader('temp.dat')

    def addOneTraceElement(self, element):
        # do not need this function in this profiler
        pass

    def run(self, mode=parda_mode.seq, *args, **kargs):
        self.c_float_array = (c_float * (self.cache_size + 1))()

        c_line_num = c_long(self.num_of_trace_elements)
        c_file_name = c_char_p(self.reader.file_loc.encode())

        if mode == parda_mode.seq:
            self.mode = mode
            self.parda_seq.classical_tree_based_stackdist(c_file_name, c_line_num, self.c_cache_size,
                                                          self.c_float_array)

        elif mode == parda_mode.openmp:
            self.num_of_threads = kargs["threads"]
            self.mode = mode
            c_threads = c_int(self.num_of_threads)
            self.parda_seq.parda_seperate_file(c_file_name, c_threads, c_line_num)
            print("all files have been separated for parallel")
            # self.parda_seq.parda_omp_stackdist.restype = POINTER(c_double)
            self.parda_seq.parda_omp_stackdist(c_file_name, c_line_num, c_threads, self.c_cache_size,
                                               self.c_float_array)

        elif mode == parda_mode.mpi:
            print('sorry, currently, mpi mode is not supported')
        elif mode == parda_mode.hybrid:
            print('sorry, currently, openmp/mpi hybrid mode is not supported')
        else:
            print("does not support given run mode")

        self.run_aux()

    def run_with_specified_lines(self, begin_line, end_line):
        c_line_num = c_long(self.num_of_trace_elements)
        c_file_name = c_char_p(self.reader.file_loc.encode())
        self.c_float_array = (c_float * (self.cache_size + 1))()

        c_begin = c_long(begin_line)
        c_end = c_long(end_line)
        self.parda_seq.classical_with_line_specification(c_file_name, c_line_num, self.c_cache_size,
                                                         c_begin, c_end, self.c_float_array)
        self.run_aux()

    def run_aux(self):
        self.calculated = True

        for i in range(self.cache_size):
            self.HRC_bin[i] = self.c_float_array[i + 1] * 100

        self.HRC[0] = self.HRC_bin[0]
        for i in range(1, self.cache_size):
            self.HRC[i] = self.HRC_bin[i] + self.HRC[i - 1]

        # the size of list is one larger than cache_size, if bin_size is 1
        self.HRC[-1] = self.HRC[-2]

        for i in range(self.num_of_blocks):
            self.MRC[i] = 100 - self.HRC[i]

        for i in range(len(self.MRC)):
            pass

    def get_reuse_distance(self):
        c_line_num = c_long(self.num_of_trace_elements)
        c_file_name = c_char_p(self.reader.file_loc.encode())
        self.c_long_array = (c_long * self.num_of_trace_elements)()

        self.parda_seq.get_reuse_distance(c_file_name, c_line_num, self.c_cache_size, self.c_long_array)
        # for i in self.c_long_array:
        #     print(i+1)

        return self.c_long_array

    def _test(self, cache_size=2000):
        c_line_num = c_long(self.num_of_trace_elements)
        c_file_name = c_char_p(self.reader.file_loc.encode())
        self.c_long_array = (c_long * (self.num_of_trace_elements + 1))()

        # self.parda_seq.main(c_line_num, c_file_name)
        self.parda_seq.get_reuse_distance_smart(c_file_name, c_line_num, self.c_cache_size, self.c_long_array)
        count = 0
        print(len(self.c_long_array))
        for i in self.c_long_array:
            # print(i)
            if -1 < i < cache_size:
                count += 1
        print(count)
        print(len(self.c_long_array))
        self.parda_seq.get_reuse_distance(c_file_name, c_line_num, self.c_cache_size, self.c_long_array)
        count = 0
        for i in self.c_long_array:
            if -1 < i < cache_size:
                count += 1
        print(count)

        def __del__(self):
            if os.path.exists('temp.dat'):
                os.remove('temp.dat')


if __name__ == "__main__":
    # p = pardaProfiler(30000, plainCacheReader("../data/parda.trace"))
    # p = pardaProfiler(30000, csvCacheReader("../data/trace_CloudPhysics_txt", 4))
    # p = parda(LRU, 3000000, basicCacheReader("temp.dat"))
    # p.run(parda_mode.seq, threads=4)
    # p.get_reuse_distance()
    import os
    import time

    # from pympler.tracker import SummaryTracker

    # tracker = SummaryTracker()

    t1 = time.time()
    for i in range(6):
        p = pardaProfiler(30000, vscsiCacheReader("../data//trace_CloudPhysics_bin"))
        # rd_a = p.get_reuse_distance()
    t2 = time.time()

    print(t2 - t1)
    p = pardaProfiler(30000, vscsiCacheReader("../data//trace_CloudPhysics_bin"))
    p.run()
    p.plotHRC()
    # print(len())
    # count = 0
    # for i in rd_a:
    #     if i==0:
    #         count+=1
    # print(count)

    # tracker.print_diff()


    # p = pardaProfiler(2000, csvCacheReader('../data/mining/mining.dat.original', 1))
    # p = pardaProfiler(2000, plainCacheReader('../data/mining/mining.dat.original'))
    # p._test()


    # for f in os.listdir('../data/mining/'):
    #     shutil.copy('../data/mining/' + f, '../data/mining/mining.dat')
    #     print(f)
    #     p._test()
    # p.run_with_specified_lines(10000, 20000)
    # p.plotHRC(autosize=True, autosize_threshhold=0.00001)



"""
this module provides the heatmap ploting engine, it supports both virtual time (fixed number of trace requests) and
real time, under both modes, it support using multiprocessing to do the plotting
currently, it only support LRU
the mechanism it works is it gets reuse distance from LRU profiler (parda for now)
"""

import logging
import os
import pickle
from collections import deque
from multiprocessing import Array

import numpy as np
from matplotlib import pyplot as plt

from mimircache.cache.LRU import LRU
from mimircache.cacheReader.vscsiReader import vscsiCacheReader
from mimircache.oldModule.pardaProfiler import pardaProfiler
from mimircache.utils.printing import *

DEBUG = True


def _calc_hit_count(reuse_dist_array, cache_size, begin_pos, end_pos, real_start):
    """

    :rtype: count of hit
    :param reuse_dist_array:
    :param cache_size:
    :param begin_pos:
    :param end_pos: end pos of trace in current partition (not included)
    :param real_start: the real start position of cache trace
    :return:
    """
    hit_count = 0
    miss_count = 0
    for i in range(begin_pos, end_pos):
        if reuse_dist_array[i] == -1:
            # never appear
            miss_count += 1
            continue
        if reuse_dist_array[i] - (i - real_start) < 0 and reuse_dist_array[i] < cache_size:
            # hit
            hit_count += 1
        else:
            # miss
            miss_count += 1
    return hit_count


def _calc_hit_rate_subprocess(order, cache_size, break_points_share_array, reuse_dist_share_array, q):
    """
    the child process for calculating hit rate, each child process will calculate for
    a column with fixed starting time
    :param order: the order of column the child is working on
    :return: a list of result in the form of (x, y, hit_rate) with x as fixed value
    """

    result_list = []
    total_hc = 0
    for i in range(order + 1, len(break_points_share_array)):
        hc = _calc_hit_count(reuse_dist_share_array, cache_size, break_points_share_array[i - 1],
                             break_points_share_array[i], break_points_share_array[order])
        total_hc += hc
        hr = total_hc / (break_points_share_array[i] - break_points_share_array[order])

        # avg_rd = 0




        result_list.append((order, i, hr))
    q.put(result_list)
    # return result_list





class heatmap:
    def __init__(self, cache_class=LRU):
        self.cache_class = cache_class


    def prepare_heatmap_dat(self, mode, reader, calculate=True):
        """
        prepare the data for plotting heatmap
        :param mode:   mode can be either virtual time(v) or real time(r)
        :param reader: for reading the trace file
        :param calculate: flag for checking whether we need to recalculate reuse distance and distribution table
        :return:
        """

        assert isinstance(reader, vscsiCacheReader), "Currently only supports vscsiReader"
        reader.reset()

        reuse_dist_python_list = []
        break_points = []

        if calculate:
            # build profiler for extracting reuse distance (For LRU)
            p = pardaProfiler(30000, reader)
            # reuse distance, c_reuse_dist_long_array is an array in dtype C long type
            c_reuse_dist_long_array = p.get_reuse_distance()

            for i in c_reuse_dist_long_array:
                reuse_dist_python_list.append(i)

            if not os.path.exists('temp/'):
                os.makedirs('temp/')
            with open('temp/reuse.dat', 'wb') as ofile:
                pickle.dump(reuse_dist_python_list, ofile)

        else:
            # the reuse distance has already been calculated, just load it
            with open('temp/reuse.dat', 'rb') as ifile:
                reuse_dist_python_list = pickle.load(ifile)

            breakpoints_filename = 'temp/break_points_' + mode + str(self.interval) + '.dat'
            # check whether the break points distribution table has been calculated
            if os.path.exists(breakpoints_filename):
                with open(breakpoints_filename, 'rb') as ifile:
                    break_points = pickle.load(ifile)

        # check break points are loaded or not, if not need to calculate it
        if not break_points:
            if mode == 'r':
                break_points = self._get_breakpoints_realtime(reader, self.interval)
            elif mode == 'v':
                break_points = self._get_breakpoints_virtualtime(self.interval, len(reuse_dist_python_list))
            with open('temp/break_points_' + mode + str(self.interval) + '.dat', 'wb') as ifile:
                pickle.dump(break_points, ifile)

        # create share memory for child process
        reuse_dist_share_array = Array('l', len(reuse_dist_python_list), lock=False)
        for i, j in enumerate(reuse_dist_python_list):
            reuse_dist_share_array[i] = j

        break_points_share_array = Array('i', len(break_points), lock=False)
        for i, j in enumerate(break_points):
            break_points_share_array[i] = j


        # create the array for storing results
        # result is a dict: (x, y) -> heat, x, y is the left, lower point of heat square
        # (x,y) means from time x to time y
        array_len = len(break_points)
        result = np.empty((array_len, array_len), dtype=np.float32)
        result[:] = np.nan

        # Jason: efficiency can be further improved by porting into Cython and improve parallel logic
        # Jason: memory usage can be optimized by not copying whole reuse distance array in each sub process
        map_list = deque()
        for i in range(len(break_points) - 1):
            map_list.append(i)





        RD_avg = []
        count_list = []
        prev = 0
        for next in break_points:
            if next==0:
                continue
            rd_avg = 0
            first_see = 0
            count = 0
            for i in range(prev, next):
                if reuse_dist_python_list[i] != -1:
                    rd_avg += reuse_dist_python_list[i]
                    count += 1
                else:
                    first_see += 1
            if count != 0:
                RD_avg.append(rd_avg/count)
            else:
                RD_avg.append(-1)
            # first_see_list.append(first_see)
            count_list.append((count, next-prev-count))
            prev = next

        for i in range(len(break_points)-1):
            for j in range(i+1, array_len):
                RD = 0
                newcount1 = 0
                newcount2 = 0
                for k in range(i,j):
                    newcount1 += count_list[i][0]   # RD != -1
                    newcount2 += count_list[i][1]
                    RD = RD + RD_avg[k]
                RD = RD/(j-i)
                # print("%8.2f"%RD)
                # if RD > (break_points[j]-self.interval//2-break_points[i]):
                #     RD -= (break_points[j]-self.interval//2-break_points[i])
                # print("%8.2f"%RD)
                hr = (break_points[j]-self.interval//2-break_points[i])/RD*self.cache_size*(newcount1/(newcount1+newcount2))
                # if hr>0.1:
                #     print(hr)
                # print('')
                result[i][j] = hr
            if i % 100 == 0:
                print(i)
                # pass




        '''
        # new 0510
        q = Queue()
        process_pool = []
        process_count = 0
        result_count = 0
        map_list_pos = 0
        while result_count < len(map_list):
            if process_count < self.num_of_process and map_list_pos < len(map_list):
                p = Process(target=_calc_hit_rate_subprocess, args=(map_list[map_list_pos], self.cache_size,
                                                                    break_points_share_array, reuse_dist_share_array,
                                                                    q))
                p.start()
                process_pool.append(p)
                process_count += 1
                map_list_pos += 1
            else:
                rl = q.get()
                for r in rl:
                    result[r[0]][r[1]] = r[2]
                process_count -= 1
                result_count += 1
            print("%2.2f%%" % (result_count / len(map_list) * 100), end='\r')
        for p in process_pool:
            p.join()
        del reuse_dist_share_array
        del break_points_share_array

        # old 0510
        # with Pool(processes=self.num_of_process) as p:
        #     for ret_list in p.imap_unordered(_calc_hit_rate_subprocess, map_list,
        #                                      chunksize=10):
        #         count += 1
        #         print("%2.2f%%" % (count / len(map_list) * 100), end='\r')
        #         for r in ret_list:  # l is a list of (x, y, hr)
        #             result[r[0]][r[1]] = r[2]

        '''
        print(result)

        with open('temp/draw', 'wb') as ofile:
            pickle.dump(result, ofile)

        return result

    @staticmethod
    def _get_breakpoints_virtualtime(bin_size, total_length):
        """

        :param bin_size: the size of the group (virtual time slice)
        :param total_length: the total length of trace requests
        :return: a list of break points
        """
        break_points = []
        for i in range(total_length // bin_size):
            break_points.append(i * bin_size)
        if break_points[-1] != total_length - 1:
            break_points.append(total_length - 1)

        if (len(break_points)) > 10000:
            colorfulPrintWithBackground('yellow', 'blue', "number of pixels in one dimension are more than 10000, \
            exact size: %d, it may take a very long time, if you didn't intend to do it, \
            please try with a larger time stamp" % len(break_points))

        return break_points

    @staticmethod
    def _get_breakpoints_realtime(reader, interval):
        """
        given reader(vscsi) and time interval, split the data according to the time interval,
        save the order into break_points(list)
        :param reader:
        :param interval:
        :return:
        """
        reader.reset()
        break_points = []
        prev = 0
        for num, line in enumerate(reader.lines()):
            if (line[0] - prev) > interval:
                break_points.append(num)
                prev = line[0]

        # noinspection PyUnboundLocalVariable
        if line[0] != prev:
            # noinspection PyUnboundLocalVariable
            break_points.append(num)

        if (len(break_points)) > 10000:
            colorfulPrintWithBackground('yellow', 'blue', "number of pixels in one dimension are more than 10000, \
            exact size: %d, it may take a very long time, if you didn't intend to do it, \
            please try with a larger time stamp" % len(break_points))

        return break_points

    @staticmethod
    def _calc_hit_rate(reuse_dist_array, cache_size, begin_pos, end_pos):
        """
        the original function for calculating hit rate, given the reuse_dist_array, cache_size,
        begin position and end position in the reuse_dist_array, return hit rate
        now isolate as a separate function
        :param reuse_dist_array:
        :param cache_size:
        :param begin_pos:
        :param end_pos:
        :return:
        """
        hit_count = 0
        miss_count = 0
        for i in range(begin_pos, end_pos):
            if reuse_dist_array[i] == -1:
                # never appear
                miss_count += 1
                continue
            if reuse_dist_array[i] - (i - begin_pos) < 0 and reuse_dist_array[i] < cache_size:
                # hit
                hit_count += 1
            else:
                # miss
                miss_count += 1
        # print("hit+miss={}, total size:{}, hit rage:{}".format(hit_count+miss_count, \
        # end_pos-begin_pos, hit_count/(end_pos-begin_pos)))
        return hit_count / (end_pos - begin_pos)

    @staticmethod
    def draw(xydict, **kargs):
        if 'figname' in kargs:
            filename = kargs['figname']
        else:
            filename = 'heatmap.png'

        plt.clf()
        masked_array = np.ma.array(xydict, mask=np.isnan(xydict))

        # print(masked_array)
        cmap = plt.cm.jet
        cmap.set_bad('w', 1.)

        # fig = plt.figure()
        # ax = fig.add_subplot(111)

        # heatmap = plt.pcolor(result2.T, cmap=plt.cm.Blues, vmin=np.min(result2[np.nonzero(result2)]), \
        # vmax=result2.max())
        try:
            # ax.pcolor(masked_array.T, cmap=cmap)
            if 'fixed_range' in kargs and kargs['fixed_range']:
                img = plt.imshow(masked_array.T, vmin=0, vmax=1, interpolation='nearest', origin='lower')
            else:
                img = plt.imshow(masked_array.T, interpolation='nearest', origin='lower')
            # cb = plt.colorbar(img, ax=ax)
            cb = plt.colorbar(img)
            if 'text' in kargs:
                (length1, length2) = masked_array.shape
                ax = plt.gca()
                # ax.text(2, 6, r'an equation: $E=mc^2$', fontsize=30)
                ax.text(length2 // 3, length1 // 8, kargs['text'], fontsize=20)  # , color='blue')

                # ax.text(length2*3//4, length1//4, 2, text, fontsize=120, color='blue')
            # plt.show()
            plt.savefig(filename)
            colorfulPrint("red", "plot is saved at the same directory")
        except Exception as e:
            logging.warning(str(e))
            plt.pcolormesh(masked_array.T, cmap=cmap)
            plt.savefig(filename)

    def __del_manual__(self):
        """
        cleaning
        :return:
        """
        if os.path.exists('temp/'):
            for filename in os.listdir('temp/'):
                os.remove('temp/' + filename)
            os.rmdir('temp/')

    def run(self, mode, interval, cache_size, reader, **kargs):
        """

        :param mode:
        :param interval:
        :param cache_size:
        :param reader:
        :param kargs: include num_of_process, figname
        :return:
        """
        self.cache_size = cache_size
        self.interval = interval
        self.mode = mode
        if 'num_of_process' in kargs:
            self.num_of_process = kargs['num_of_process']
        else:
            self.num_of_process = 4


        if 'calculate' in kargs:
            calculate = kargs['calculate']
        else:
            calculate = True

        if mode == 'r':
            xydict = self.prepare_heatmap_dat('r', reader, calculate)
        elif mode == 'v':
            xydict = self.prepare_heatmap_dat('v', reader, calculate)
        else:
            raise RuntimeError("Cannot recognize this mode, it can only be either real time(r) or virtual time(v), "
                               "but you input %s" % mode)

        self.draw(xydict, **kargs)

        # self.__del_manual__()


def server_plot_all():
    mem_sizes = []
    with open('memSize', 'r') as ifile:
        for line in ifile:
            mem_sizes.append(int(line.strip()))

    for filename in os.listdir("../data/cloudphysics"):
        print(filename)
        if filename.endswith('.vscsitrace'):
            if int(filename.split('_')[0][1:]) in [1, 4, 5, 51, 99, 83, 87]:
                continue
            if os.path.exists(filename + '_v.png'):
                continue
            hm = heatmap()
            mem_size = mem_sizes[int(filename.split('_')[0][1:]) - 1] * 16
            reader = vscsiCacheReader("../data/cloudphysics/" + filename)
            hm.run('v', 100000, mem_size, reader, num_of_process=45, figname=filename + '_v.png')


def server_size_plot():
    mem_sizes = []
    with open('memSize', 'r') as ifile:
        for line in ifile:
            mem_sizes.append(int(line.strip()))

    for filename in os.listdir("../data/cloudphysics"):
        if filename.endswith('.vscsitrace'):
            if int(filename.split('_')[0][1:]) in [1, 4, 5, 51, 99, 83, 87]:
                continue
            hm = heatmap()
            mem_size = mem_sizes[int(filename.split('_')[0][1:])]
            reader = vscsiCacheReader("../data/cloudphysics/" + filename)
            print(filename)
            size = 512
            while size < mem_size * 1024 * 128:
                print(size)
                if size == 512:
                    hm.run('r', 1000000000, size, reader, num_of_process=42, figname='time_size/' + filename
                                                                                     + '_' + str(size) + '.png',
                           calculate=True, fixed_range=True, text='size=' + str(size))
                else:
                    hm.run('r', 1000000000, size, reader, num_of_process=42, figname='time_size/' + filename
                                                                                     + '_' + str(size) + '.png',
                           calculate=False, fixed_range=True, text='size=' + str(size))
                size *= 2


def localtest():
    # reader1 = plainCacheReader("../data/parda.trace")
    reader2 = vscsiCacheReader("../data/trace_CloudPhysics_bin")

    hm = heatmap()
    hm.run('v', 100, 2000, reader2, num_of_process=4, fixed_range="False", text="Hello word")





if __name__ == "__main__":
    localtest()
    # server_plot_all()



from mimircache.cache.adaptiveSLRU import AdaptiveSLRU
from mimircache.cache.FIFO import FIFO
from mimircache.cache.LFU_MRU import LFU_MRU
from mimircache.cache.LFU_RR import LFU_RR
from mimircache.cache.LRU import LRU
from mimircache.cache.MRU import MRU
from mimircache.cache.Random import Random
from mimircache.cache.S4LRU import S4LRU
from mimircache.cache.SLRU import SLRU
from mimircache.cache.clock import clock
from mimircache.cacheReader.csvReader import csvCacheReader
from mimircache.cacheReader.plainReader import plainCacheReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader
from mimircache.profiler.LRUProfiler import LRUProfiler
from mimircache.profiler.generalProfiler import generalProfiler
from mimircache.profiler.heatmap import heatmap
from mimircache.profiler.cGeneralProfiler import cGeneralProfiler
from mimircache.profiler.cHeatmap import cHeatmap
from mimircache.const import *
from mimircache.profiler.twoDPlots import *


class cachecow:
    def __init__(self, **kwargs):
        self.reader = None
        self.cacheclass_mapping = {}
        self.prepare_cacheclass_mapping()

    def prepare_cacheclass_mapping(self):
        self.cacheclass_mapping['lru'] = LRU
        self.cacheclass_mapping['arc'] = AdaptiveSLRU
        self.cacheclass_mapping['clock'] = clock
        self.cacheclass_mapping['fifo'] = FIFO
        self.cacheclass_mapping['lfu_mru'] = LFU_MRU
        self.cacheclass_mapping['lfu_rr'] = LFU_RR
        self.cacheclass_mapping['mru'] = MRU
        self.cacheclass_mapping['random'] = Random
        self.cacheclass_mapping['slru'] = SLRU
        self.cacheclass_mapping['s4lru'] = S4LRU

    def open(self, file_path):
        # assert os.path.exists(file_path), "data file does not exist"
        self.reader = plainCacheReader(file_path)
        return self.reader

    def csv(self, file_path, column):
        self.reader = csvCacheReader(file_path, column=column)
        return self.reader

    def vscsi(self, file_path):
        self.reader = vscsiCacheReader(file_path)
        return self.reader

    def set_size(self, size):
        assert isinstance(size, int), "size can only be an integer"
        self.cache_size = size

    def reset(self):
        self.reader.reset()

    def _profiler_pre_check(self, **kwargs):
        """
        check whether user has provided new cache size and data information
        :param kwargs:
        :return:
        """
        reader = None

        if 'num_of_process' in kwargs:
            num_of_process = kwargs['num_of_process']
        else:
            num_of_process = DEFAULT_NUM_OF_PROCESS

        if 'data' in kwargs and 'dataType' in kwargs:
            if kwargs['dataType'] == 'plain':
                reader = plainCacheReader(kwargs['data'])
            if kwargs['dataType'] == 'csv':
                assert 'column' in kwargs, "you didn't provide column number for csv reader"
                reader = csvCacheReader(kwargs['data'], kwargs['column'])
            if kwargs['dataType'] == 'vscsi':
                reader = vscsiCacheReader(kwargs['data'])
        elif 'reader' in kwargs:
            reader = kwargs['reader']
        else:
            reader = self.reader

        assert reader, "you didn't provide a reader nor data (data file and data type)"
        self.reader = reader

        return reader, num_of_process

    def heatmap(self, mode, interval, plot_type, algorithm="LRU", cache_params=None, cache_size=-1, **kwargs):
        """

        :param mode:
        :param interval:
        :param kwargs: algorithm:
        :return:
        """

        reader, num_of_process = self._profiler_pre_check(**kwargs)
        if algorithm.lower() in c_available_cache:
            hm = cHeatmap()
            print("cache size : " + str(cache_size))
            hm.heatmap(reader, mode, interval, plot_type,
                       cache_size=cache_size,
                       algorithm=cache_alg_mapping[algorithm.lower()],
                       cache_params=cache_params,
                       num_of_threads=num_of_process,
                       **kwargs)
        else:
            hm = heatmap()
            hm.heatmap(reader, mode, interval, plot_type,
                       cache_size=cache_size,
                       algorithm=cache_alg_mapping[algorithm.lower()],
                       cache_params=cache_params,
                       **kwargs)

    def differential_heatmap(self, mode, interval, plot_type, algorithm1,
                             algorithm2="Optimal", cache_params1=None, cache_params2=None, cache_size=-1, **kwargs):
        """
        alg2 - alg1
        :param alg1:
        :param alg2:
        :param mode:
        :param interval:
        :param plot_type:
        :param kwargs:
        :return:
        """
        figname = 'differential_heatmap.png'
        if 'figname' in kwargs:
            figname = kwargs['figname']

        assert cache_size != -1, "you didn't provide size for cache"

        reader, num_of_process = self._profiler_pre_check(**kwargs)

        if algorithm1.lower() in c_available_cache and algorithm2.lower() in c_available_cache:
            hm = cHeatmap()
            hm.differential_heatmap(reader, mode, interval, plot_type,
                                    cache_size=cache_size,
                                    algorithm1=cache_alg_mapping[algorithm1.lower()],
                                    algorithm2=cache_alg_mapping[algorithm2.lower()],
                                    cache_params1=cache_params1,
                                    cache_params2=cache_params2,
                                    num_of_threads=num_of_process,
                                    **kwargs)

        else:
            hm = heatmap()
            if algorithm1.lower() not in c_available_cache:
                xydict1 = hm.calculate_heatmap_dat(reader, mode, interval, plot_type,
                                                   cache_size=cache_size, algorithm=algorithm1,
                                                   cache_params=cache_params1,
                                                   **kwargs)[0]
            else:
                xydict1 = c_heatmap.heatmap(reader.cReader, mode, interval, plot_type,
                                            cache_size=cache_size, algorithm=algorithm1,
                                            cache_params=cache_params1,
                                            num_of_threads=num_of_process)

            if algorithm2.lower() not in c_available_cache:
                xydict2 = hm.calculate_heatmap_dat(reader, mode, interval, plot_type,
                                                   cache_size=cache_size, algorithm=algorithm2,
                                                   cache_params=cache_params2,
                                                   **kwargs)[0]
            else:
                xydict2 = c_heatmap.heatmap(reader.cReader, mode, interval, plot_type,
                                            cache_size=cache_size, algorithm=algorithm2,
                                            cache_params=cache_params2,
                                            num_of_threads=num_of_process)

            cHm = cHeatmap()
            text = "      differential heatmap\n      cache size: {},\n      cache type: {}-{},\n" \
                   "      time type: {},\n      time interval: {},\n      plot type: \n{}".format(
                cache_size, algorithm2, algorithm1, mode, interval, plot_type)

            x1, y1 = xydict1.shape
            x1 = int(x1 / 2.8)
            y1 /= 8
            if mode == 'r':
                cHm.set_plot_params('x', 'real_time', xydict=xydict1, label='start time (real)',
                                    text=(x1, y1, text))
                cHm.set_plot_params('y', 'real_time', xydict=xydict1, label='end time (real)')
            else:
                cHm.set_plot_params('x', 'virtual_time', xydict=xydict1, label='start time (virtual)',
                                    text=(x1, y1, text))
                cHm.set_plot_params('y', 'virtual_time', xydict=xydict1, label='end time (virtual)')

            cHm.draw_heatmap((xydict2 - xydict1) / xydict1, figname=figname)

    def profiler(self, algorithm, cache_params=None, cache_size=-1, **kwargs):
        """
        profiler
        :param cache_class:
        :param kwargs:
        :return:
        """
        reader, num_of_process = self._profiler_pre_check(**kwargs)

        profiler = None
        cache_params = None
        bin_size = -1

        if algorithm.lower() == "lru":
            profiler = LRUProfiler(reader, cache_size)
        else:
            assert cache_size != -1, "you didn't provide size for cache"
            if 'bin_size' in kwargs:
                bin_size = kwargs['bin_size']

            if isinstance(algorithm, str):
                if algorithm.lower() in c_available_cache:
                    profiler = cGeneralProfiler(reader, algorithm, cache_size, bin_size, cache_params, num_of_process)
                else:
                    profiler = generalProfiler(reader, self.cacheclass_mapping[algorithm.lower()], cache_size, bin_size,
                                               cache_params, num_of_process)
            else:
                profiler = generalProfiler(reader, algorithm, cache_size, bin_size, cache_params, num_of_process)

        return profiler

    def __iter__(self):
        assert self.reader, "you haven't provided a data file"
        return self.reader

    def next(self):
        return self.__next__()

    def __next__(self):  # Python 3
        return self.reader.next()

    def __del__(self):
        if self.reader:
            self.reader.close()

    def twoDPlot(self, mode, time_interval, plot_type):
        if plot_type == 'cold_miss':
            cold_miss_2d(self.reader, mode, time_interval)
        elif plot_type == 'request_num':
            request_num_2d(self.reader, mode, time_interval)
        else:
            print("currently don't support your specified plot_type: " + str(plot_type))



if __name__ == "__main__":
    CACHE_SIZE = 2000
    c = cachecow()
    c.vscsi('../data/trace.vscsi')
    # m.heatmap('r', 100000000)

    # m.test()
    # m.open('../data/parda.trace')
    p = c.profiler("LRU")
    print(p.get_reuse_distance())
    p.plotHRC()

    p = c.profiler('optimal', cache_size=CACHE_SIZE)
    print(p.get_hit_count())

    c.twoDPlot('v', 1000, "cold_miss")

    c.heatmap('r', 1000000, "hit_rate_start_time_end_time", "LRU", cache_size=CACHE_SIZE, num_of_process=8,
              figname="abc.png")

    c.differential_heatmap('r', 1000000, "hit_rate_start_time_end_time", "LRU", cache_size=CACHE_SIZE, num_of_process=8,
                           figname='2.png')

    c.differential_heatmap('r', 10000000, "hit_rate_start_time_end_time", "LRU", "ARC", cache_size=CACHE_SIZE,
                           num_of_process=8, figname='3.png')

    # c.differential_heatmap('r', 1000000, "hit_rate_start_time_end_time", "LRU", num_of_process=8, figname='2.png')

    TIME_INTERVAL = 10000000000
    from mimircache.cacheReader.vscsiReader import vscsiCacheReader

    PATH = '/run/shm/traces/'
    c = cachecow()

    for f in os.listdir(PATH):
        if f.endswith('vscsitrace'):
            # reader = vscsiCacheReader(PATH + f)
            c.vscsi(PATH + f)
            c.heatmap('r', TIME_INTERVAL, "rd_distribution",
                      figname='0627_rd_distribution/' + f + '_rd_distribution_' + str(TIME_INTERVAL) + '.png')
            # request_num_2d(reader, 'r', TIME_INTERVAL,
            #                figname='0627_rd_distribution/' + f + '_rd_distribution_' + str(TIME_INTERVAL) + '.png')



from mimircache.cache.abstractCache import cache


class MRU(cache):
    def __init__(self, cache_size=1000):
        super(MRU, self).__init__(cache_size)
        self.cacheDict = dict()
        self.last_element = None

    def checkElement(self, element):
        """
        :param element:
        :return: whether the given element is in the cache
        """
        if element in self.cacheDict:
            return True
        else:
            return False

    def _updateElement(self, element):
        """ the given element is in the cache, now its frequency
        :param element:
        :return: original rank
        """
        self.last_element = element

    def _insertElement(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: True on success, False on failure
        """
        if len(self.cacheDict) == self.cache_size:
            self._evictOneElement()
        self.cacheDict[element] = element
        self.last_element = element

    def find_evict_key(self):
        return self.last_element

    def _printCacheLine(self):
        for key, value in self.cacheDict.items():
            print("{}: {}".format(key, value))

    def _evictOneElement(self):
        """
        evict one element from the cache line
        :return: True on success, False on failure
        """
        evict_key = self.find_evict_key()
        del self.cacheDict[evict_key]

    def addElement(self, element):
        """
        :param element: the element in the reference, it can be in the cache, or not
        :return: -1 if not in cache, otherwise old rank
        """
        # print(element, end=': \t')
        if self.checkElement(element):
            self._updateElement(element)
            return True
        else:
            self._insertElement(element)
            return False

    def __repr__(self):
        return "MRU, given size: {}, current size: {}".format(self.cache_size, len(self.cacheDict))

    def __str__(self):
        return self.__repr__()



""" this module is used for all other cache replacement algorithms excluding LRU(LRU also works, but slow compared to
    using pardaProfiler),
"""
# -*- coding: utf-8 -*-


import math
import os

# deal with headless situation
import traceback

import matplotlib
# matplotlib.use('Agg')

from multiprocessing import Process, Pipe, Array
from os import path

from mimircache.cacheReader.vscsiReader import vscsiCacheReader

from mimircache.cache.adaptiveSLRU import AdaptiveSLRU
from mimircache.cache.clock import clock
from mimircache.cache.FIFO import FIFO
from mimircache.cache.LFU_LRU__NEED_OPTIMIZATION import LFU_LRU
from mimircache.cache.LFU_MRU import LFU_MRU
from mimircache.cache.LFU_RR import LFU_RR
from mimircache.cache.LRU import LRU
from mimircache.cache.MRU import MRU
from mimircache.cache.Random import Random
from mimircache.cache.SLRU import SLRU
from mimircache.cache.S4LRU import S4LRU
from mimircache.cache.Optimal import optimal
import matplotlib.ticker as ticker

from mimircache.cacheReader.plainReader import plainCacheReader

import mimircache.c_generalProfiler as c_generalProfiler

import matplotlib.pyplot as plt

from mimircache.const import c_available_cache, cache_alg_mapping
from mimircache.profiler.abstract.abstractProfiler import profilerAbstract
from mimircache.const import *


class cGeneralProfiler:
    def __init__(self, reader, cache_name, cache_size, bin_size=-1, cache_params=None,
                 num_of_threads=DEFAULT_NUM_OF_PROCESS):
        assert cache_name.lower() in cache_alg_mapping, "please check your cache replacement algorithm: " + cache_name
        assert cache_name.lower() in c_available_cache, \
            "cGeneralProfiler currently only available on the following caches: {}\n, " \
            "please use generalProfiler".format(c_available_cache)

        self.reader = reader
        self.cache_size = cache_size
        self.cache_name = cache_alg_mapping[cache_name.lower()]
        if bin_size == -1:
            self.bin_size = int(self.cache_size / DEFAULT_BIN_NUM_PROFILER)
        else:
            self.bin_size = bin_size
        self.cache_params = cache_params
        self.num_of_threads = num_of_threads

        # if the given file is not basic reader, needs conversion
        need_convert = True
        for instance in c_available_cacheReader:
            if isinstance(reader, instance):
                need_convert = False
                break
        if need_convert:
            self.prepare_file()

    all = ["get_hit_count", "get_hit_rate", "get_miss_rate", "plotMRC", "plotHRC"]

    def prepare_file(self):
        self.num_of_lines = 0
        with open('temp.dat', 'w') as ofile:
            i = self.reader.read_one_element()
            while i is not None:
                self.num_of_lines += 1
                ofile.write(str(i) + '\n')
                i = self.reader.read_one_element()
        self.reader = plainCacheReader('temp.dat')


    def get_hit_count(self, **kwargs):
        """

        :return: a numpy array, with hit count corresponding to size [0, bin_size, bin_size*2 ...]
        """
        sanity_kwargs = {}
        if 'num_of_threads' not in kwargs:
            sanity_kwargs['num_of_threads'] = self.num_of_threads
        if 'cache_size' in kwargs:
            cache_size = kwargs['cache_size']
        else:
            cache_size = self.cache_size
        if 'begin' in kwargs:
            sanity_kwargs['begin'] = kwargs['begin']
        if 'end' in kwargs:
            sanity_kwargs['end'] = kwargs['end']
        return c_generalProfiler.get_hit_count(self.reader.cReader, self.cache_name, cache_size,
                                               self.bin_size, cache_params=self.cache_params, **sanity_kwargs)

    def get_hit_rate(self, **kwargs):
        """

        :return: a numpy array, with hit rate corresponding to size [0, bin_size, bin_size*2 ...]
        """
        sanity_kwargs = {}
        if 'num_of_threads' not in kwargs:
            sanity_kwargs['num_of_threads'] = self.num_of_threads
        if 'cache_size' in kwargs:
            cache_size = kwargs['cache_size']
        else:
            cache_size = self.cache_size
        if 'begin' in kwargs:
            sanity_kwargs['begin'] = kwargs['begin']
        if 'end' in kwargs:
            sanity_kwargs['end'] = kwargs['end']
        return c_generalProfiler.get_hit_rate(self.reader.cReader, self.cache_name, cache_size,
                                              self.bin_size, cache_params=self.cache_params, **sanity_kwargs)

    def get_miss_rate(self, **kwargs):
        """

        :return: a numpy array, with miss rate corresponding to size [0, bin_size, bin_size*2 ...]
        """

        sanity_kwargs = {}
        if 'num_of_threads' not in kwargs:
            sanity_kwargs['num_of_threads'] = self.num_of_threads
        if 'cache_size' in kwargs:
            cache_size = kwargs['cache_size']
        else:
            cache_size = self.cache_size
        if 'begin' in kwargs:
            sanity_kwargs['begin'] = kwargs['begin']
        if 'end' in kwargs:
            sanity_kwargs['end'] = kwargs['end']
        return c_generalProfiler.get_miss_rate(self.reader.cReader, self.cache_name, cache_size,
                                               self.bin_size, cache_params=self.cache_params, **sanity_kwargs)

    def plotMRC(self, figname="MRC.png", **kwargs):
        MRC = self.get_miss_rate(**kwargs)
        try:
            # tick = ticker.FuncFormatter(lambda x, pos: '{:2.0f}'.format(x * self.bin_size))
            # plt.gca().xaxis.set_major_formatter(tick)
            plt.xlim(0, self.cache_size)
            plt.plot(range(0, self.cache_size + 1, self.bin_size), MRC)
            plt.xlabel("cache Size")
            plt.ylabel("Miss Rate")
            plt.title('Miss Rate Curve', fontsize=18, color='black')
            plt.savefig(figname, dpi=600)
            plt.show()
            plt.clf()
            del MRC
        except Exception as e:
            plt.savefig(figname)
            print("the plotting function is not wrong, is this a headless server?")
            print(e)

    def plotHRC(self, figname="HRC.png", **kwargs):
        HRC = self.get_hit_rate(**kwargs)
        try:
            # tick = ticker.FuncFormatter(lambda x, pos: '{:2.0f}'.format(x * self.bin_size))
            # plt.gca().xaxis.set_major_formatter(tick)

            plt.xlim(0, self.cache_size)
            plt.plot(range(0, self.cache_size + 1, self.bin_size), HRC)
            plt.xlabel("cache Size")
            plt.ylabel("Hit Rate")
            plt.title('Hit Rate Curve', fontsize=18, color='black')
            plt.savefig(figname, dpi=600)
            plt.show()
            plt.clf()
            del HRC
        except Exception as e:
            plt.savefig(figname)
            print("the plotting function is not wrong, is this a headless server?")
            print(e)


    def __del__(self):
        if (os.path.exists('temp.dat')):
            os.remove('temp.dat')


def server_plot_all(path='/run/shm/traces/', threads=48):
    from mimircache.profiler.LRUProfiler import LRUProfiler
    folder = '0628_HRC'
    for filename in os.listdir(path):
        print(filename)
        if filename.endswith('.vscsitrace'):
            reader = vscsiCacheReader(path + filename)
            p1 = LRUProfiler(reader)
            size = p1.plotHRC(figname=folder + "/" + filename + '_LRU_HRC.png')

            if os.path.exists(folder + "/" + filename + '_LRU4_HRC_' + str(size) + '_.png'):
                reader.close()
                continue

            p2 = cGeneralProfiler(reader, 'Optimal', cache_size=size, bin_size=int(size / 500), num_of_threads=threads)
            p2.plotHRC(figname=folder + "/" + filename + '_Optimal_HRC_' + str(size) + '_.png')
            p3 = cGeneralProfiler(reader, "LRU_K", cache_size=size, cache_params={"K": 2}, bin_size=int(size / 500))
            p3.plotHRC(figname=folder + "/" + filename + '_LRU2_HRC_' + str(size) + '_.png', num_of_threads=threads)
            # p4 = cGeneralProfiler(reader, "LRU_K", cache_size=size, cache_params={"K": 3}, bin_size=int(size / 500))
            # p4.plotHRC(figname=folder + "/" + filename + '_LRU3_HRC_' + str(size) + '_.png', num_of_threads=threads)
            # p5 = cGeneralProfiler(reader, "LRU_K", cache_size=size, cache_params={"K": 4}, bin_size=int(size / 500))
            # p5.plotHRC(figname=folder + "/" + filename + '_LRU4_HRC_' + str(size) + '_.png', num_of_threads=threads)
            reader.close()





if __name__ == "__main__":
    import time

    server_plot_all('../data/', threads=8)

    # t1 = time.time()
    # r = plainCacheReader('../../data/test')
    # r = plainCacheReader('../data/parda.trace')
    r = vscsiCacheReader('../data/trace.vscsi')
    # cg = cGeneralProfiler(r, 'Optimal', 2000, 200)
    cg = cGeneralProfiler(r, 'LRU_2', 2000, 200)

    t1 = time.time()

    # print(cg.get_hit_rate())
    # print(cg.get_hit_count())
    # print(cg.get_miss_rate())

    t2 = time.time()
    print("TIME: %f" % (t2 - t1))

    # print(cg.plotMRC())
    # print(cg.get_hit_rate(begin=1, end=20))



from mimircache.cache.LRU import LRU
from mimircache.cache.abstractCache import cache


class SLRU(cache):
    def __init__(self, cache_size=1000, ratio=1, *args, **kwargs):
        """

        :param cache_size: size of cache
        :param args: raio: the ratio of protected/probationary
        :return:
        """
        super().__init__(cache_size)
        self.ratio = ratio
        # Maybe use two linkedlist and a dict will be more efficient?
        self.protected = LRU(int(self.cache_size * self.ratio / (self.ratio + 1)))
        self.probationary = LRU(int(self.cache_size * 1 / (self.ratio + 1)))

    def checkElement(self, element):
        """
        :param element:
        :return: whether the given element is in the cache
        """
        if element in self.protected or element in self.probationary:
            return True
        else:
            return False

    def _updateElement(self, element):
        """ the given element is in the cache, now update it to new location
        :param element:
        :return: None
        """
        if element in self.protected:
            self.protected._updateElement(element)
        else:
            # element is in probationary, remove from probationary, insert to end of protected,
            # evict from protected to probationary if needed

            # get the node and remove from probationary
            node = self.probationary.cacheDict[element]
            self.probationary.cacheLinkedList.removeNode(node)
            del self.probationary.cacheDict[element]

            # insert into protected
            evicted_key = self.protected._insertElement(node.content)

            # if there are element evicted from protected area, add to probationary area
            if evicted_key:
                self.probationary._insertElement(evicted_key)

    def _insertElement(self, element):
        """
        the given element is not in the cache, now insert it into cache
        :param element:
        :return: evicted element
        """
        return self.probationary._insertElement(element)

    def _printCacheLine(self):
        print("protected: ")
        self.protected._printCacheLine()
        print("probationary: ")
        self.probationary._printCacheLine()

    def _evictOneElement(self):
        """
        evict one element from the cache line
        :return: True on success, False on failure
        """
        pass

    def addElement(self, element):
        """
        :param element: a cache request, it can be in the cache, or not
        :return: None
        """
        if self.checkElement(element):
            self._updateElement(element)
            # self.printCacheLine()
            return True
        else:
            self._insertElement(element)
            # self.printCacheLine()
            return False

    def __repr__(self):
        return "SLRU, given size: {}, given protected part size: {}, given probationary part size: {}, \
            current protected part size: {}, current probationary size: {}". \
            format(self.cache_size, self.protected.cache_size, self.probationary.cache_size,
                   self.protected.cacheLinkedList.size, self.probationary.cacheLinkedList.size)



from mimircache.cacheReader.abstractReader import cacheReaderAbstract
import mimircache.c_cacheReader as c_cacheReader


class plainCacheReader(cacheReaderAbstract):
    def __init__(self, file_loc):
        super(plainCacheReader, self).__init__(file_loc)
        # cacheReaderAbstract.__init__(self, file_loc)
        self.trace_file = open(file_loc, 'r')
        self.cReader = c_cacheReader.setup_reader(file_loc, 'p')

    def read_one_element(self):
        super().read_one_element()
        line = self.trace_file.readline()
        if line:
            return line.strip()
        else:
            return None

    def __next__(self):  # Python 3
        super().__next__()
        element = self.trace_file.readline().strip()
        if element:
            return element
        else:
            raise StopIteration

    def __repr__(self):
        return "basic cache reader, cache trace separated by line, %s" % super().__repr__()


if __name__ == "__main__":
    reader = plainCacheReader('../data/trace_CloudPhysics')

    # usage one: for reading all elements
    for i in reader:
        print(i)

    # usage two: best for reading one element each time
    s = reader.read_one_element()
    # s2 = next(reader)
    while s:
        print(s)
        s = reader.read_one_element()
        # s2 = next(reader)



import os
import configparser
from mimircache.cacheReader.csvReader import csvCacheReader
from mimircache.cacheReader.plainReader import plainCacheReader
from mimircache.cacheReader.vscsiReader import vscsiCacheReader

# global c_available_cache
c_available_cache = []
c_available_cacheReader = [plainCacheReader, vscsiCacheReader, csvCacheReader]
cache_alg_mapping = {}
BASE_DIR = os.path.dirname(__file__)

DEFAULT_BIN_NUM_PROFILER = 100
DEFAULT_NUM_OF_PROCESS = 4


def init():
    _init_C_available_modules()
    _init_cache_alg_mapping()


def _init_C_available_modules():
    config = configparser.ConfigParser()
    # print(BASE_DIR + '/conf')
    config.read(BASE_DIR + '/conf')
    if 'C_available_cache' in config.sections():
        c_available_cache.extend(config['C_available_cache'])
    else:
        raise RuntimeWarning("cannot find any cache module in C")

        # if 'C_available_reader' in config.sections():
        #     c_available_cacheReader.extend(config['C_available_reader'])
        # else:
        #     raise RuntimeWarning("cannot find any cacheReader module in C")





def _init_cache_alg_mapping():
    """
    match all possible cache replacement algorithm names(lower case) to available cache replacement algorithms
    :return:
    """

    cache_alg_mapping['optimal'] = 'Optimal'
    cache_alg_mapping['rr'] = "Random"
    cache_alg_mapping['lru'] = "LRU"
    cache_alg_mapping['fifo'] = "FIFO"
    cache_alg_mapping['arc'] = "ARC"
    cache_alg_mapping['clock'] = "clock"
    cache_alg_mapping['mru'] = "MRU"
    cache_alg_mapping['slru'] = "SLRU"
    cache_alg_mapping['s4lru'] = "S4LRU"
    cache_alg_mapping['lfu_rr'] = "LFU_RR"
    cache_alg_mapping['lfu_mru'] = "LFU_MRU"
    cache_alg_mapping['lru_k'] = "LRU_K"
    cache_alg_mapping['lru_2'] = "LRU_2"
